"""
Training Configuration UI

This module provides a graphical interface for configuring AI training parameters.
"""

import pygame
import json
import os
from typing import Dict, Any, Optional, Tuple, List

class ConfigUI:
    def __init__(self, width: int = 800, height: int = 600):
        """Initialize the configuration UI."""
        pygame.init()
        self.width = width
        self.height = height
        self.screen = pygame.display.set_mode((width, height))
        pygame.display.set_caption("Training Configuration")
        
        # Fonts
        self.title_font = pygame.font.Font(None, 48)
        self.header_font = pygame.font.Font(None, 36)
        self.text_font = pygame.font.Font(None, 24)
        
        # Colors
        self.colors = {
            'background': (0, 0, 0),
            'text': (200, 200, 200),
            'highlight': (0, 255, 0),
            'button': (40, 40, 40),
            'button_hover': (60, 60, 60),
            'input_bg': (30, 30, 30),
            'input_active': (50, 50, 50)
        }
        
        # Parameters
        self.parameters = {
            'timesteps': {
                'value': 1000000,
                'type': 'int',
                'min': 1000,
                'max': 10000000,
                'description': 'Total timesteps to train for'
            },
            'learning_rate': {
                'value': 0.0003,
                'type': 'float',
                'min': 0.00001,
                'max': 0.01,
                'description': 'Learning rate for training'
            },
            'batch_size': {
                'value': 64,
                'type': 'int',
                'min': 32,
                'max': 512,
                'description': 'Batch size for training'
            },
            'n_envs': {
                'value': 8,
                'type': 'int',
                'min': 1,
                'max': 32,
                'description': 'Number of parallel environments'
            },
            'n_steps': {
                'value': 2048,
                'type': 'int',
                'min': 128,
                'max': 8192,
                'description': 'Number of steps per update'
            }
        }
        
        # UI state
        self.active_input = None
        self.input_text = ""
        self.scroll_offset = 0
        self.max_scroll = max(0, len(self.parameters) * 60 - (height - 200))
        
        # Load saved config if exists
        self.config_file = "training_config.json"
        self.load_config()
    
    def load_config(self) -> None:
        """Load configuration from file."""
        if os.path.exists(self.config_file):
            try:
                with open(self.config_file, 'r') as f:
                    saved_config = json.load(f)
                for key, value in saved_config.items():
                    if key in self.parameters:
                        self.parameters[key]['value'] = value
            except:
                pass
    
    def save_config(self) -> None:
        """Save configuration to file."""
        config = {key: param['value'] for key, param in self.parameters.items()}
        with open(self.config_file, 'w') as f:
            json.dump(config, f, indent=4)
    
    def draw_text_input(self, rect: pygame.Rect, value: Any, active: bool) -> None:
        """Draw a text input field."""
        color = self.colors['input_active'] if active else self.colors['input_bg']
        pygame.draw.rect(self.screen, color, rect)
        pygame.draw.rect(self.screen, self.colors['text'], rect, 1)
        
        text = str(value)
        if active:
            text = self.input_text + "|"
        
        text_surface = self.text_font.render(text, True, self.colors['text'])
        text_rect = text_surface.get_rect(midleft=(rect.left + 5, rect.centery))
        self.screen.blit(text_surface, text_rect)
    
    def draw_button(self, rect: pygame.Rect, text: str, hover: bool = False) -> None:
        """Draw a button."""
        color = self.colors['button_hover'] if hover else self.colors['button']
        pygame.draw.rect(self.screen, color, rect)
        pygame.draw.rect(self.screen, self.colors['text'], rect, 1)
        
        text_surface = self.text_font.render(text, True, self.colors['text'])
        text_rect = text_surface.get_rect(center=rect.center)
        self.screen.blit(text_surface, text_rect)
    
    def validate_input(self, param: Dict[str, Any], value: str) -> Optional[Any]:
        """Validate and convert input value."""
        try:
            if param['type'] == 'int':
                val = int(value)
            else:
                val = float(value)
            
            if val < param['min'] or val > param['max']:
                return None
            return val
        except:
            return None
    
    def run(self) -> Optional[Dict[str, Any]]:
        """Run the configuration UI. Returns the config dict if saved, None if cancelled."""
        running = True
        save_clicked = False
        mouse_pos = (0, 0)
        
        while running:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    running = False
                
                elif event.type == pygame.MOUSEBUTTONDOWN:
                    mouse_pos = event.pos
                    
                    # Check parameter inputs
                    y = 100 - self.scroll_offset
                    for name, param in self.parameters.items():
                        input_rect = pygame.Rect(300, y, 200, 30)
                        if input_rect.collidepoint(mouse_pos):
                            self.active_input = name
                            self.input_text = str(param['value'])
                        y += 60
                    
                    # Check buttons
                    save_rect = pygame.Rect(self.width//2 - 150, self.height - 60, 140, 40)
                    cancel_rect = pygame.Rect(self.width//2 + 10, self.height - 60, 140, 40)
                    
                    if save_rect.collidepoint(mouse_pos):
                        self.save_config()
                        save_clicked = True
                        running = False
                    elif cancel_rect.collidepoint(mouse_pos):
                        running = False
                
                elif event.type == pygame.MOUSEBUTTONUP:
                    if event.button == 4:  # Mouse wheel up
                        self.scroll_offset = max(0, self.scroll_offset - 30)
                    elif event.button == 5:  # Mouse wheel down
                        self.scroll_offset = min(self.max_scroll, self.scroll_offset + 30)
                
                elif event.type == pygame.KEYDOWN:
                    if self.active_input is not None:
                        if event.key == pygame.K_RETURN:
                            param = self.parameters[self.active_input]
                            if val := self.validate_input(param, self.input_text):
                                param['value'] = val
                            self.active_input = None
                        elif event.key == pygame.K_BACKSPACE:
                            self.input_text = self.input_text[:-1]
                        else:
                            if event.unicode.isnumeric() or event.unicode == '.':
                                self.input_text += event.unicode
            
            # Draw UI
            self.screen.fill(self.colors['background'])
            
            # Title
            title = "Training Configuration"
            title_surface = self.title_font.render(title, True, self.colors['highlight'])
            title_rect = title_surface.get_rect(midtop=(self.width//2, 20))
            self.screen.blit(title_surface, title_rect)
            
            # Parameters
            y = 100 - self.scroll_offset
            for name, param in self.parameters.items():
                if 0 <= y <= self.height - 100:
                    # Parameter name and description
                    name_surface = self.header_font.render(name, True, self.colors['text'])
                    desc_surface = self.text_font.render(param['description'], True, self.colors['text'])
                    self.screen.blit(name_surface, (20, y))
                    self.screen.blit(desc_surface, (20, y + 30))
                    
                    # Input field
                    input_rect = pygame.Rect(300, y, 200, 30)
                    self.draw_text_input(input_rect, param['value'], name == self.active_input)
                y += 60
            
            # Buttons
            save_rect = pygame.Rect(self.width//2 - 150, self.height - 60, 140, 40)
            cancel_rect = pygame.Rect(self.width//2 + 10, self.height - 60, 140, 40)
            
            save_hover = save_rect.collidepoint(mouse_pos)
            cancel_hover = cancel_rect.collidepoint(mouse_pos)
            
            self.draw_button(save_rect, "Save", save_hover)
            self.draw_button(cancel_rect, "Cancel", cancel_hover)
            
            pygame.display.flip()
        
        pygame.quit()
        
        if save_clicked:
            return {key: param['value'] for key, param in self.parameters.items()}
        return None 