"""
Snake Game Environment for Reinforcement Learning

This module provides a gym-like interface for training AI agents to play the snake game.
It includes:
- State observation space
- Action space
- Reward system
- Environment dynamics
"""

import numpy as np
import random
import gymnasium as gym
from gymnasium import spaces
from typing import Tuple, List, Dict, Any, Optional
import pygame
from src.core import Snake
from src.core import Direction
from src.core import GameSession
from src.ui import GameArea
from src.config import GameRules

def manhattan_distance(pos1: Tuple[int, int], pos2: Tuple[int, int]) -> int:
    """Calculate Manhattan distance between two points."""
    return abs(pos1[0] - pos2[0]) + abs(pos1[1] - pos2[1])

def manhattan_distance_wrap(pos1: Tuple[int, int], pos2: Tuple[int, int], size: int) -> int:
    """Calculate Manhattan distance between two points with wrap-around."""
    dx = abs(pos1[0] - pos2[0])
    dy = abs(pos1[1] - pos2[1])
    return min(dx, size - dx) + min(dy, size - dy)

class SnakeEnv(gym.Env):
    """A Gymnasium environment for training snake AI agents."""
    
    metadata = {"render_modes": ["human", "rgb_array"], "render_fps": 30}
    
    def __init__(self, size=30, render_mode=None, difficulty="easy"):
        super().__init__()
        self.size = size
        self.render_mode = render_mode
        self.difficulty = difficulty
        
        # Action to Direction mapping
        self.action_to_direction = {
            0: Direction.UP,
            1: Direction.RIGHT,
            2: Direction.DOWN,
            3: Direction.LEFT
        }
        
        # Direction to index mapping for one-hot encoding
        self.direction_to_index = {
            Direction.UP: 0,
            Direction.RIGHT: 1,
            Direction.DOWN: 2,
            Direction.LEFT: 3
        }
        
        # Initialize reward scales with emphasis on precision
        self.reward_scales = {
            'food': 20.0,           # Base food reward (will be scaled exponentially)
            'death': -5.0,          # Reduced death penalty to encourage exploration
            'distance': 0.5,        # Increased distance reward
            'survival': -0.01,      # Small survival penalty
            'milestone': 10.0,      # Larger milestone rewards
            'efficiency': 0.2,      # Small efficiency reward
            'exploration': 0.5,     # Increased exploration reward
            'safety': 1.0,         # Safety reward for avoiding danger
            'timeout': -1.0,       # Reduced timeout penalty
            'wrap_bonus': 0.2,     # Small wrap-around bonus
            'near_miss': -0.1,            # Very small near miss penalty
            'repetitive': -2      # Increased base penalty for repetitive movement
        }
        
        # Near miss detection
        self.near_miss_threshold = 1  # Only penalize very close misses
        
        # Time limit parameters - more lenient
        self.base_time_limit = 200     # Double base time limit
        self.length_time_bonus = 100    # Double length bonus
        self.max_time_limit = 1000     # Much longer maximum time
        
        # Exploration decay
        self.initial_exploration_bonus = 2.0   # Higher initial exploration
        self.exploration_decay = 0.9995       # Slower decay
        self.min_exploration_bonus = 0.5      # Higher minimum exploration
        self.current_exploration_bonus = self.initial_exploration_bonus
        self.total_episodes = 0             # Track number of episodes for decay
        
        # Curriculum learning parameters
        self.curriculum_stage = 0
        self.success_threshold = {
            0: 3,    # Stage 0: Basic movement (need more consistent success)
            1: 5,    # Stage 1: Food collection
            2: 7,    # Stage 2: Longer snake
            3: 10    # Stage 3: Full difficulty
        }
        self.consecutive_successes = 0
        self.stage_requirements = {
            0: 1,    # Stage 0: Get one food
            1: 3,    # Stage 1: Get three food
            2: 5,    # Stage 2: Get five food
            3: 7     # Stage 3: Get seven food
        }
        
        # Dynamic episode length based on curriculum stage
        self.base_max_steps = 300  # More steps for exploration
        self.max_steps = self.base_max_steps
        
        # Window dimensions for rendering
        self.window_width = 1024
        self.window_height = 768
        
        # Create game area and rules
        self.game_area = GameArea(self.window_width, self.window_height)
        self.rules = GameRules()
        self.rules.wrap_around = False  # No wrap-around for AI training
        
        # Define action and observation spaces
        self.action_space = spaces.Discrete(4)  # Up, Right, Down, Left
        
        num_channels = 26  # Total number of channels we created in _get_normalized_observation
        self.observation_space = spaces.Box(
            low=0,
            high=1,
            shape=(num_channels, size, size),  # (channels, height, width)
            dtype=np.float32
        )
        
        # Game session
        self.session = None
        self.steps = 0
        self.steps_since_food = 0
        self.current_time = 0
        self.last_score = 0
        self.last_distance = float('inf')
        self.last_food_steps = 0
        self.steps_since_direction_change = 0  # Track steps since last direction change
        self.last_direction = None  # Track previous direction
        self.recent_positions = []
        self.steps_in_same_direction = 0  # Track steps without direction change
        self.last_min_food_distance = float('inf')  # Initialize minimum distance tracking
        
        # For rendering
        self.window = None
        self.clock = None
        
        self.score_milestone_rewards = {5: 10.0, 10: 20.0, 15: 30.0}  # Bonus rewards at score milestones
        self.last_milestone = 0  # Track last milestone reached
        self.base_survival_penalty = -1.0  # Much larger constant penalty per step
        
        # Movement control parameters
        self.max_direction_steps = 10       # Reduced from 20 to 10
        self.direction_step_increase = 2    # Reduced from 5 to 2
        self.min_safe_turns = 2            # Reduced from 3 to 2
        self.repetitive_threshold = 10     # Base threshold before repetitive penalty kicks in
        self.repetitive_scale = 1.2        # Base scale for penalty growth
        self.repetitive_wrap_scale = 2.0   # Much faster growth rate in wrap mode
        self.max_repetitive_penalty = -5.0  # Increased maximum repetitive movement penalty
        
        # Initialize base rewards with adjusted values
        rewards = {
            'food': 0.0,
            'death': 0.0,
            'distance': 0.0,
            'survival': self.base_survival_penalty * 0.5,
            'direction': 0.0,
            'efficiency': 0.0,
            'progress': 0.0,
            'alignment': 0.0,
            'repetitive': 0.0
        }
    
    def _get_normalized_observation(self) -> np.ndarray:
        """
        Create a multi-channel 2D observation from the game state.
        Spatial channels encode the grid layout (snake & food),
        while extra scalar features are broadcast over additional channels.
        """
        state = self.session.get_state()
        snake_body = state["snake_body"]
        food_position = state["food_position"]
        head = snake_body[0]
        body = snake_body

        # --- Spatial Channels ---
        # Channel 0: Snake positions (1 for snake, 0 otherwise)
        snake_channel = np.zeros((self.size, self.size), dtype=np.float32)
        for pos in snake_body:
            snake_channel[pos[1], pos[0]] = 1.0
            assert pos[0] < self.size and pos[1] < self.size, \
                f"Snake position {pos} exceeds grid size {self.size}"

        # Channel 1: Food position
        food_channel = np.zeros((self.size, self.size), dtype=np.float32)
        food_channel[food_position[1], food_position[0]] = 1.0

        # --- Auxiliary Scalar Features ---
        # Compute distances and apply wrap-around adjustments
        dx = food_position[0] - head[0]
        dy = food_position[1] - head[1]
        if self.rules.wrap_around:
            if dx > self.size/2:
                wrap_dx = dx - self.size
            elif dx < -self.size/2:
                wrap_dx = dx + self.size
            else:
                wrap_dx = dx
            if dy > self.size/2:
                wrap_dy = dy - self.size
            elif dy < -self.size/2:
                wrap_dy = dy + self.size
            else:
                wrap_dy = dy
        else:
            wrap_dx = dx
            wrap_dy = dy

        # Normalize
        dx_norm = dx / self.size
        dy_norm = dy / self.size
        wrap_dx_norm = wrap_dx / self.size
        wrap_dy_norm = wrap_dy / self.size

        # Danger indicators (vector of 4 values)
        dangers = self._get_danger_observations_all_directions(head, body)
        
        # Other auxiliary scalar features
        snake_length_norm = len(body) / (self.size**2)
        # One-hot encode current direction
        direction_one_hot = [0.0] * 4
        direction_one_hot[self.direction_to_index[state["snake_direction"]]] = 1.0

        # One-hot encode relative food direction
        food_dir = [0.0] * 4
        if abs(dx_norm) > abs(dy_norm):
            if dx_norm > 0: 
                food_dir[self.direction_to_index[Direction.RIGHT]] = 1.0
            else: 
                food_dir[self.direction_to_index[Direction.LEFT]] = 1.0
        else:
            if dy_norm > 0: 
                food_dir[self.direction_to_index[Direction.DOWN]] = 1.0
            else: 
                food_dir[self.direction_to_index[Direction.UP]] = 1.0

        current_time_limit = min(
            self.max_time_limit,
            self.base_time_limit + (len(body) - 1) * self.length_time_bonus
        )
        time_pressure = np.clip(self.steps_since_food / current_time_limit, 0, 1)
        wrap_mode = float(self.rules.wrap_around)
        progress = min(1.0, len(body) / 10)
        recent_food = min(1.0, self.steps_since_food / 50)
        curriculum_stage_norm = float(self.curriculum_stage) / 3
        advanced_length = float(len(body) > 5)
        manhattan_norm = min(1.0, manhattan_distance(head, food_position) / (self.size * 2))

        # --- Create "broadcasted" channels for each scalar ---
        def broadcast_channel(value):
            return np.full((self.size, self.size), value, dtype=np.float32)
        
        dx_channel         = broadcast_channel(dx_norm)
        dy_channel         = broadcast_channel(dy_norm)
        wrap_dx_channel    = broadcast_channel(wrap_dx_norm)
        wrap_dy_channel    = broadcast_channel(wrap_dy_norm)
        snake_length_chan  = broadcast_channel(snake_length_norm)
        time_pressure_chan = broadcast_channel(time_pressure)
        wrap_mode_chan     = broadcast_channel(wrap_mode)
        progress_chan      = broadcast_channel(progress)
        recent_food_chan   = broadcast_channel(recent_food)
        curriculum_chan    = broadcast_channel(curriculum_stage_norm)
        advanced_length_chan = broadcast_channel(advanced_length)
        manhattan_chan     = broadcast_channel(manhattan_norm)

        # For one-hot features, create one channel per value
        direction_channels = [broadcast_channel(v) for v in direction_one_hot]
        food_dir_channels = [broadcast_channel(v) for v in food_dir]

        # For dangers, one channel per direction (4 channels)
        danger_channels = [broadcast_channel(d) for d in dangers]

        # --- Stack all channels ---
        # You can adjust the channel order as desired.
        channels = [
            snake_channel,        # Channel 0
            food_channel,         # Channel 1
            dx_channel,           # Channel 2
            dy_channel,           # Channel 3
            wrap_dx_channel,      # Channel 4
            wrap_dy_channel,      # Channel 5
        ]
        channels.extend(danger_channels)            # Channels 6-9
        channels.append(snake_length_chan)            # Channel 10
        channels.extend(direction_channels)           # Channels 11-14
        channels.extend(food_dir_channels)            # Channels 15-18
        channels.append(time_pressure_chan)           # Channel 19
        channels.append(wrap_mode_chan)               # Channel 20
        channels.append(progress_chan)                # Channel 21
        channels.append(recent_food_chan)             # Channel 22
        channels.append(curriculum_chan)              # Channel 23
        channels.append(advanced_length_chan)         # Channel 24
        channels.append(manhattan_chan)               # Channel 25

        # Final observation: shape (num_channels, size, size)
        observation = np.stack(channels, axis=0)
        return np.clip(observation, 0, 1)

    def _get_danger_observations_all_directions(self, head: Tuple[int, int], body: List[Tuple[int, int]]) -> List[float]:
        """Get danger observations in all four directions."""
        dangers = [0.0] * 4  # [UP, RIGHT, DOWN, LEFT]
        
        # Check each direction
        directions = [Direction.UP, Direction.RIGHT, Direction.DOWN, Direction.LEFT]
        for i, direction in enumerate(directions):
            dx, dy = direction.to_vector()
            next_pos = (head[0] + dx, head[1] + dy)
            
            # Check wall collision
            if not self.rules.wrap_around:
                if (next_pos[0] < 0 or next_pos[0] >= self.size or 
                    next_pos[1] < 0 or next_pos[1] >= self.size):
                    dangers[i] = 1.0
                    continue
            else:
                next_pos = (next_pos[0] % self.size, next_pos[1] % self.size)
            
            # Check self collision
            if next_pos in body[1:]:
                dangers[i] = 1.0
        
        return dangers
    
    def reset(self, seed=None, options=None):
        """Reset with curriculum-based difficulty and exploration decay."""
        # Call parent reset without unpacking
        super().reset(seed=seed)
        if seed is not None:
            random.seed(seed)
            np.random.seed(seed)
        
        # Update exploration bonus
        self.total_episodes += 1
        self.current_exploration_bonus = max(
            self.min_exploration_bonus,
            self.initial_exploration_bonus * (self.exploration_decay ** self.total_episodes)
        )
        
        # Initialize game session with curriculum-appropriate settings
        self.session = GameSession(
            self.size,
            self.size,
            self.rules
        )
        
        # Reset counters
        self.steps = 0
        self.steps_since_food = 0
        self.current_time = 0
        self.last_score = 0
        self.last_distance = float('inf')
        self.last_direction = self.session.get_state()["snake_direction"]  # Track initial direction
        self.recent_positions = []
        self.steps_in_same_direction = 0  # Track steps without direction change
        self.last_min_food_distance = float('inf')  # Initialize minimum distance tracking
        
        # Get initial observation
        observation = self._get_normalized_observation()
        info = {
            "curriculum_stage": self.curriculum_stage,
            "exploration_bonus": self.current_exploration_bonus
        }
        
        return observation, info
    
    def _calculate_area_penalty(self, head_pos):
        """Calculate penalty for staying in a small area."""
        # Add current position to history
        self.recent_positions.append(head_pos)
        if len(self.recent_positions) > self.max_recent_positions:
            self.recent_positions.pop(0)
        
        if len(self.recent_positions) < 10:  # Need minimum history to calculate
            return 0.0
        
        # Calculate bounding box of recent positions
        x_coords = [p[0] for p in self.recent_positions]
        y_coords = [p[1] for p in self.recent_positions]
        area_width = max(x_coords) - min(x_coords) + 1
        area_height = max(y_coords) - min(y_coords) + 1
        area = area_width * area_height
        
        # Calculate unique positions visited
        unique_positions = len(set(self.recent_positions))
        
        # Penalize small areas and repeated positions
        area_penalty = 0.0
        if area < 9:  # 3x3 grid or smaller
            area_penalty = -0.2 * (9 - area) / 8  # Max penalty -0.2 for 1x1 area
        
        # Additional penalty for revisiting same positions frequently
        repetition_penalty = -0.3 * (1 - unique_positions / len(self.recent_positions))
        
        return area_penalty + repetition_penalty
    
    def step(self, action):
        """Improved reward structure with length-aware exploration."""
        self.steps += 1
        self.steps_since_food += 1
        
        # Calculate current time limit based on snake length
        snake_length = len(self.session.get_state()["snake_body"])
        current_time_limit = min(
            self.max_time_limit,
            self.base_time_limit + (snake_length - 1) * self.length_time_bonus
        )
        
        # Get state before action
        prev_state = self.session.get_state()
        prev_head = prev_state["snake_head"]
        prev_food_dist = manhattan_distance(prev_head, prev_state["food_position"])
        if self.rules.wrap_around:
            prev_food_dist = min(prev_food_dist, 
                               manhattan_distance_wrap(prev_head, prev_state["food_position"], self.size))
        
        # Convert action to Direction and apply
        direction = self.action_to_direction[int(action)]
        
        # Update game state
        state, base_reward, done = self.session.step(direction, self.current_time)
        
        # Initialize reward components
        rewards = {
            'food': 0.0,
            'death': 0.0,
            'distance': 0.0,
            'survival': 0.0,
            'milestone': 0.0,
            'efficiency': 0.0,
            'exploration': 0.0,
            'safety': 0.0,
            'timeout': 0.0,
            'direction': 0.0,
            'repetitive': 0.0
        }
        
        # Get current state info
        curr_head = state["snake_body"][0]
        curr_food = state["food_position"]
        curr_food_dist = manhattan_distance(curr_head, curr_food)
        if self.rules.wrap_around:
            curr_food_dist = min(curr_food_dist, 
                               manhattan_distance_wrap(curr_head, curr_food, self.size))

        # Update current time
        self.current_time += self.session.move_cooldown
        
        # Food reward with exponential scaling based on snake length
        food_eaten = state["score"] > prev_state["score"]
        if food_eaten:
            # Reset timer when food is eaten
            self.steps_since_food = 0
            
            # Calculate exponential food reward
            base_food_reward = self.reward_scales['food']
            length_multiplier = 1.2 ** (len(state["snake_body"]) - 1)  # 20% increase per length
            rewards['food'] = base_food_reward * length_multiplier
            
            # Add milestone rewards for achieving certain scores
            if state["score"] in [5, 10, 15, 20]:
                rewards['milestone'] = self.reward_scales['milestone'] * (state["score"] / 5)
        else:
            # Only apply survival penalty if we haven't eaten food in a while
            if self.steps_since_food > current_time_limit / 2:
                penalty_factor = (self.steps_since_food - current_time_limit/2) / (current_time_limit/2)
                rewards['survival'] = self.reward_scales['survival'] * penalty_factor
        
        # Check if direction changed and update counters
        direction_changed = direction != self.last_direction
        if direction_changed:
            self.steps_in_same_direction = 0
        else:
            self.steps_in_same_direction += 1
            # Apply growing penalty for repetitive movement
            if self.steps_in_same_direction > (self.repetitive_threshold - 5 if self.rules.wrap_around else self.repetitive_threshold):
                excess_steps = self.steps_in_same_direction - (self.repetitive_threshold - 5 if self.rules.wrap_around else self.repetitive_threshold)
                # Use much faster growth rate in wrap-around mode
                growth_rate = self.repetitive_wrap_scale if self.rules.wrap_around else self.repetitive_scale
                repetitive_penalty = self.reward_scales['repetitive'] * (growth_rate ** excess_steps)
                
                # Apply additional multiplier in wrap-around mode that grows with steps
                if self.rules.wrap_around:
                    wrap_multiplier = 1.0 + (excess_steps * 0.5)  # Multiplier grows with each step
                    repetitive_penalty *= wrap_multiplier
                    rewards['repetitive'] = repetitive_penalty  # No maximum cap - let it grow unbounded in wrap-around mode

                else:
                    # Cap the penalty when not in wrap-around mode
                    rewards['repetitive'] = max(repetitive_penalty, self.max_repetitive_penalty)
                
        
        self.last_direction = direction
        
        # Check for timeout or stuck in loop - only if we haven't just eaten food
        if (self.steps_since_food >= current_time_limit and not food_eaten):
            done = True
            rewards['timeout'] = self.reward_scales['timeout']
            # Scale timeout penalty based on distance to food
            if curr_food_dist < 5:  # Harsher penalty for timing out near food
                rewards['timeout'] *= 1.5
            # Additional penalty for getting stuck in a loop
            if self.steps_in_same_direction >= 30:  # Also update this threshold
                rewards['timeout'] *= 1.2
        
        # Death penalty - calculate AFTER food rewards
        if done and not self.steps_since_food >= current_time_limit: 
            base_death_penalty = self.reward_scales['death']
            
            # Check if death was due to self-collision
            if curr_head in state["snake_body"][1:]:
                # Base multiplier for self-collision
                self_collision_multiplier = 2.0
                
                # Additional penalty scaling with score
                # At score 5: 2.5x penalty
                # At score 10: 3.0x penalty
                # At score 15: 3.5x penalty
                score_multiplier = 1.0 + (state["score"] / 10)
                
                # Apply both multipliers
                rewards['death'] = base_death_penalty * self_collision_multiplier * score_multiplier
                
                # Additional penalty if we died near food
                if curr_food_dist < 5:
                    rewards['death'] *= 1.5  # Even harsher if we collide with ourselves near food
            else:
                # For wall collisions, keep original scaling but make it less punishing at higher scores
                rewards['death'] = base_death_penalty * (1.0 - min(0.5, state["score"] / 20))
                
            # Add an immediate negative reward for losing potential score
            potential_loss_penalty = -0.5 * state["score"]  # Larger penalty for dying with higher scores
            rewards['death'] += potential_loss_penalty
        
        # Track minimum distance to food and check for near misses
        if curr_food_dist < self.last_min_food_distance:
            self.last_min_food_distance = curr_food_dist
        elif curr_food_dist > self.last_min_food_distance:
            # We're moving away from our closest approach
            if self.last_min_food_distance <= self.near_miss_threshold:
                # We got very close but missed
                rewards['distance'] += self.reward_scales['near_miss']
            self.last_min_food_distance = float('inf')  # Reset tracking
        
        # Progressive distance reward - more reward for getting closer when near food
        distance_change = prev_food_dist - curr_food_dist
        if curr_food_dist < 5:  # When close to food
            distance_multiplier = 2.0  # Double the reward/penalty
        else:
            distance_multiplier = 1.0
        rewards['distance'] = distance_change * self.reward_scales['distance'] * distance_multiplier
        
        # Calculate direction penalty
        if curr_food_dist <= 5:  # Only apply when close to food
            # Calculate optimal direction to food
            dx = curr_food[0] - curr_head[0]
            dy = curr_food[1] - curr_head[1]
            
            if self.rules.wrap_around:
                # Adjust for wrap-around
                if dx > self.size/2: dx -= self.size
                elif dx < -self.size/2: dx += self.size
                if dy > self.size/2: dy -= self.size
                elif dy < -self.size/2: dy += self.size
            
            # Determine optimal direction(s)
            optimal_directions = []
            if abs(dx) > abs(dy):
                if dx > 0: optimal_directions.append(Direction.RIGHT)
                elif dx < 0: optimal_directions.append(Direction.LEFT)
            if abs(dy) >= abs(dx):
                if dy > 0: optimal_directions.append(Direction.DOWN)
                elif dy < 0: optimal_directions.append(Direction.UP)
            
            # Base penalty scales with proximity (closer = higher penalty)
            base_penalty = (6 - curr_food_dist) * 0.1  # Scales from 0.1 to 0.5
            
            if direction not in optimal_directions:
                # Check if we're moving directly away from food
                opposite_dirs = {
                    Direction.UP: Direction.DOWN,
                    Direction.DOWN: Direction.UP,
                    Direction.LEFT: Direction.RIGHT,
                    Direction.RIGHT: Direction.LEFT
                }
                optimal_opposites = [opposite_dirs[d] for d in optimal_directions]
                
                if direction in optimal_opposites:
                    # Double penalty for moving directly away from food
                    rewards['direction'] = -2 * base_penalty
                else:
                    # Regular penalty for suboptimal direction
                    rewards['direction'] = -base_penalty
            elif curr_food_dist <= 2:  # Small reward for correct direction when very close
                rewards['direction'] = 0.1
        
        # Efficiency reward - encourage purposeful movement
        if not done and self.steps_since_food < 50:  # Only apply when actively hunting
            rewards['efficiency'] = self.reward_scales['efficiency'] * (1.0 - self.steps_since_food / 50)
        
        # Calculate final reward - simple sum of components
        reward = sum(rewards.values())
        
        # Get observation
        observation = self._get_normalized_observation()
        
        # No more episode-based truncation
        truncated = False
        
        # Include reward components and exploration info in info dict for monitoring
        info = {
            "score": state["score"],
            "reward_components": rewards,
            "exploration_bonus": self.current_exploration_bonus,
            "direction_changed": direction_changed,
            "steps_in_same_direction": self.steps_in_same_direction,
            "snake_length": snake_length,
            "steps": self.steps  # Include total steps for monitoring
        }
        
        return observation, reward, done, truncated, info
    
    def render(self):
        """Render the environment."""
        if self.window is None and self.render_mode == "human":
            pygame.init()
            self.window = pygame.display.set_mode((self.window_width, self.window_height))
            pygame.display.set_caption("Snake Environment")
            self.clock = pygame.time.Clock()
        
        if self.window is not None:
            self.window.fill((0, 0, 0))
            
            # Draw game area
            self.game_area.draw(self.window)
            
            # Draw game session
            self.session.render(self.window, self.game_area)
            
            # Draw score
            font = pygame.font.Font(None, 36)
            score_text = f"Score: {self.session.score}"
            score_surface = font.render(score_text, True, (0, 255, 0))
            score_rect = score_surface.get_rect(midtop=(self.window_width // 2, 20))
            self.window.blit(score_surface, score_rect)
            
            pygame.display.flip()
            self.clock.tick(self.metadata["render_fps"])
        
        if self.render_mode == "rgb_array":
            return np.transpose(
                np.array(pygame.surfarray.pixels3d(self.window)),
                axes=(1, 0, 2)
            )
    
    def close(self):
        """Close the environment."""
        if self.window is not None:
            pygame.quit()
            self.window = None

    def get_state(self) -> Dict:
        """Get the current game state."""
        state = self.session.get_state()
        state['steps_since_food'] = self.steps_since_food
        
        # Calculate and include current time limit
        snake_length = len(state["snake_body"])
        current_time_limit = min(
            self.max_time_limit,
            self.base_time_limit + (snake_length - 1) * self.length_time_bonus
        )
        state['current_time_limit'] = current_time_limit
        
        return state

    def _get_adjacent_positions(self, position: Tuple[int, int]) -> List[Tuple[int, int]]:
        """
        Get all adjacent positions to the given position.
        
        Args:
            position: Current position (x, y)
            
        Returns:
            List of adjacent positions [(x, y), ...]
            
        Raises:
            ValueError: If position is None or not a tuple of 2 integers
            ValueError: If position coordinates are not integers
        """
        if not position or not isinstance(position, tuple) or len(position) != 2:
            raise ValueError("Position must be a tuple of 2 coordinates")
            
        try:
            x, y = int(position[0]), int(position[1])
        except (TypeError, ValueError):
            raise ValueError("Position coordinates must be integers")
            
        # Get positions in all 4 directions (up, right, down, left)
        adjacent = [
            (x, y - 1),  # up
            (x + 1, y),  # right 
            (x, y + 1),  # down
            (x - 1, y)   # left
        ]
        
        # Validate size attribute exists and is positive integer
        if not hasattr(self, 'size') or not isinstance(self.size, int) or self.size <= 0:
            raise ValueError("Environment size must be a positive integer")
            
        # If wrap-around is enabled, adjust out-of-bounds positions
        if hasattr(self, 'rules') and hasattr(self.rules, 'wrap_around') and self.rules.wrap_around:
            adjacent = [
                (pos[0] % self.size, pos[1] % self.size)
                for pos in adjacent
            ]
            return adjacent
        
        # Otherwise, only return positions that are within bounds
        return [
            (pos[0], pos[1]) for pos in adjacent
            if 0 <= pos[0] < self.size and 0 <= pos[1] < self.size
        ]

    def _update_curriculum(self, score: int, done: bool) -> None:
        """Update curriculum stage based on agent's performance."""
        if done:
            if score >= self.stage_requirements[self.curriculum_stage]:
                self.consecutive_successes += 1
            else:
                self.consecutive_successes = 0
            
            # Check for stage advancement
            if (self.curriculum_stage < 3 and 
                self.consecutive_successes >= self.success_threshold[self.curriculum_stage]):
                self.curriculum_stage += 1
                self.consecutive_successes = 0
                # Increase episode length with curriculum stage
                self.max_steps = self.base_max_steps * (1 + self.curriculum_stage * 0.5)
                
    def _get_curriculum_rules(self) -> GameRules:
        """Get game rules appropriate for current curriculum stage."""
        rules = GameRules()
        
        # Adjust rules based on curriculum stage
        if self.curriculum_stage == 0:
            rules.wrap_around = True  # Make it easier to avoid walls
            rules.speed_increase = False
            rules.initial_move_cooldown = 200  # Much slower movement
        elif self.curriculum_stage == 1:
            rules.wrap_around = True
            rules.speed_increase = False  # Still no speed increase
            rules.initial_move_cooldown = 150
        elif self.curriculum_stage == 2:
            rules.wrap_around = True  # Keep wrap-around until final stage
            rules.speed_increase = True
            rules.initial_move_cooldown = 120
        else:  # Stage 3
            rules.wrap_around = False
            rules.speed_increase = True
            rules.initial_move_cooldown = 100
        
        return rules

# Register the environment with Gymnasium
from gymnasium.envs.registration import register

register(
    id='Snake-v0',
    entry_point='src.ai.environment:SnakeEnv',
    kwargs={
        'size': 30,
        'render_mode': None,
        'difficulty': 'easy'
    }
) 