import gym
import numpy as np
import torch
import torch.nn as nn
from stable_baselines3.common.torch_layers import BaseFeaturesExtractor


class CustomRecurrentCNN(BaseFeaturesExtractor):
    """
    A simple CNN with an LSTM layer added on top. This enables the network to keep an internal
    memory of past observations so that it can adapt to changing rules/environment dynamics.
    """
    def __init__(self, observation_space, features_dim=512):
        # Initialize with a features_dim that matches the LSTM's output.
        super(CustomRecurrentCNN, self).__init__(observation_space, features_dim)
        
        # Calculate the expected CNN output size first
        n_input_channels = observation_space.shape[0]  # Number of input channels
        
        self.cnn = nn.Sequential(
            # Layer 1: (n_channels, 30, 30) -> (32, 15, 15)
            nn.Conv2d(n_input_channels, 32, kernel_size=3, stride=2, padding=1),
            nn.ReLU(),
            
            # Layer 2: (32, 15, 15) -> (64, 8, 8)
            nn.Conv2d(32, 64, kernel_size=3, stride=2, padding=1),
            nn.ReLU(),
            
            # Layer 3: (64, 8, 8) -> (64, 4, 4)
            nn.Conv2d(64, 64, kernel_size=3, stride=2, padding=1),
            nn.ReLU(),
            
            nn.Flatten()
        )
        
        # Calculate CNN output dimension using a dummy forward pass
        with torch.no_grad():
            dummy_input = torch.zeros(1, *observation_space.shape)
            cnn_output = self.cnn(dummy_input)
            self._cnn_output_dim = cnn_output.shape[1]  # Should be 64 * 4 * 4 = 1024
        
        # Add linear layer to reduce CNN output to desired feature dimension
        self.fc = nn.Linear(self._cnn_output_dim, 256)
        
        # LSTM layer
        self.lstm = nn.LSTM(256, features_dim, batch_first=True)

    def forward(self, observations):
        # Pass through CNN: (batch, channels, height, width) -> (batch, cnn_features)
        cnn_out = self.cnn(observations)
        
        # Pass through linear layer: (batch, cnn_features) -> (batch, 256)
        fc_out = self.fc(cnn_out)
        
        # Add time dimension for LSTM: (batch, 256) -> (batch, 1, 256)
        lstm_input = fc_out.unsqueeze(1)
        
        # LSTM: (batch, 1, 256) -> (batch, 1, features_dim)
        lstm_out, _ = self.lstm(lstm_input)
        
        # Remove time dimension: (batch, 1, features_dim) -> (batch, features_dim)
        return lstm_out.squeeze(1)


class CustomCNN(BaseFeaturesExtractor):
    """
    Custom feature extractor for the snake environment.
    Uses fully connected layers since our input is a 1D vector of normalized features.
    """
    
    def __init__(self, observation_space: gym.spaces.Box, features_dim: int = 512):
        super().__init__(observation_space, features_dim)
        self.input_dim = int(np.prod(observation_space.shape))
        
        self.cnn = nn.Sequential(
            nn.Linear(self.input_dim, 512),
            nn.ReLU(),
            nn.BatchNorm1d(512),
            nn.Linear(512, 512),
            nn.ReLU(),
            nn.BatchNorm1d(512),
            nn.Linear(512, features_dim),
            nn.ReLU(),
        )
    
    def forward(self, observations: torch.Tensor) -> torch.Tensor:
        # Handle both single and batch observations
        if len(observations.shape) == 1:
            observations = observations.unsqueeze(0)
        
        # Ensure the input dimension matches what we expect
        flat_obs = observations.reshape(observations.shape[0], self.input_dim)
        return self.cnn(flat_obs)