"""
Training script for Snake AI using Stable Baselines3.

This script sets up and trains RL agents using the SnakeEnv environment.
It supports multiple difficulty levels and provides real-time visualization
of the training process.
"""

import os
import time
import numpy as np
import torch
from stable_baselines3 import PPO
from stable_baselines3.common.vec_env import DummyVecEnv, SubprocVecEnv
from stable_baselines3.common.callbacks import BaseCallback
from stable_baselines3.common.utils import set_random_seed
from stable_baselines3.common.monitor import Monitor
import gymnasium as gym
import torch.nn as nn
import argparse
from queue import Queue
from threading import Thread
from typing import Dict, Any, Optional
import json
from queue import Empty

from src.ai.environment import SnakeEnv
from src.ai.visualize import run_dashboard
from src.ai.model_networks import CustomRecurrentCNN, CustomCNN

class VisualizationCallback(BaseCallback):
    """Callback for updating the training visualization."""
    
    def __init__(self, viz_queue: Queue, total_timesteps: int, eval_freq: int = 1000, n_steps: int = 512, n_envs: int = 32, initial_timesteps: int = 0):
        super().__init__()
        self.viz_queue = viz_queue
        self.total_timesteps = total_timesteps
        self.eval_freq = eval_freq
        self.n_steps = n_steps
        self.n_envs = n_envs
        self.initial_timesteps = initial_timesteps  # Store initial timesteps
        
        # Training metrics
        self.episode_rewards = []
        self.episode_lengths = []
        
        # Evaluation metrics
        self.eval_scores = []
        self.high_score = 0
        self.recent_high_score = 0
        
        # Current episode tracking
        self.current_reward = 0
        self.current_length = 0
        
        # Evaluation games (one with wrap, one without)
        self.eval_envs = [
            Monitor(SnakeEnv(render_mode="rgb_array")) for _ in range(2)
        ]
        self.eval_envs[0].unwrapped.rules.wrap_around = False
        self.eval_envs[1].unwrapped.rules.wrap_around = True
        
        self.eval_obs = [None, None]
        self.eval_done = [True, True]
        self.eval_scores_current = [0, 0]
        
        # Real-time demo environment
        self.demo_env = Monitor(SnakeEnv(render_mode="rgb_array"))
        self.demo_env.unwrapped.rules.wrap_around = True
        self.demo_obs = None
        self.demo_done = True
        self.last_demo_time = 0
        self.demo_speed = 50
    
    def _on_step(self) -> bool:
        """Update visualization on each step."""
        try:
            current_time = time.time() * 1000  # Convert to milliseconds
            
            # Update episode tracking
            self.current_length += 1
            reward = self.locals.get("rewards", [0])[0]
            self.current_reward += reward
            
            # Get action probabilities and value estimate for demo game
            if self.demo_obs is not None:
                # Ensure proper tensor shape (batch_size, obs_dim)
                demo_obs_tensor = torch.as_tensor(self.demo_obs).float()
                if len(demo_obs_tensor.shape) == 1:
                    demo_obs_tensor = demo_obs_tensor.unsqueeze(0)
                
                # Move tensor to same device as model's policy
                device = self.model.policy.device
                demo_obs_tensor = demo_obs_tensor.to(device)
                
                with torch.no_grad():
                    # Get action distribution and value estimate
                    dist = self.model.policy.get_distribution(demo_obs_tensor)
                    action_probs = dist.distribution.probs
                    value_estimate = self.model.policy.predict_values(demo_obs_tensor)
                
                # Move results back to CPU and get first item (since we added batch dimension)
                action_probs = action_probs[0].cpu().numpy()
                value_estimate = value_estimate[0].cpu().numpy()
            else:
                action_probs = np.zeros(4)
                value_estimate = np.array([0.0])
            
            # Check if episode ended
            dones = self.locals.get("dones", [False])
            if any(dones):
                self.episode_rewards.append(self.current_reward)
                self.episode_lengths.append(self.current_length)
                self.current_reward = 0
                self.current_length = 0
            
            # Run evaluation games (keep these deterministic as they're meant to show best performance)
            for i, (env, obs, done) in enumerate(zip(self.eval_envs, self.eval_obs, self.eval_done)):
                if done:
                    if obs is not None:  # If this wasn't the first reset
                        self.eval_scores.append(self.eval_scores_current[i])
                        # Update high scores based on evaluation performance
                        self.high_score = max(self.high_score, self.eval_scores_current[i])
                        self.recent_high_score = max(self.recent_high_score, self.eval_scores_current[i])
                    self.eval_obs[i] = env.reset()[0]
                    self.eval_done[i] = False
                    self.eval_scores_current[i] = 0
                
                eval_action, _ = self.model.predict(self.eval_obs[i], deterministic=True)
                eval_action = eval_action.item() if hasattr(eval_action, 'item') else eval_action
                self.eval_obs[i], _, done, truncated, info = env.step(eval_action)
                self.eval_done[i] = done or truncated
                # Update current evaluation score
                self.eval_scores_current[i] = info.get("score", 0)
            
            # Run real-time demo game
            if current_time - self.last_demo_time >= self.demo_speed:
                if self.demo_done:
                    self.demo_obs = self.demo_env.reset()[0]
                    self.demo_done = False
                
                demo_action, _ = self.model.predict(self.demo_obs, deterministic=True)
                demo_action = demo_action.item() if hasattr(demo_action, 'item') else demo_action
                self.demo_obs, _, done, truncated, info = self.demo_env.step(demo_action)
                self.demo_done = done or truncated
                self.last_demo_time = current_time
            
            # Get current states from actual training environments
            # Sample one no-wrap and one wrap environment from the training environments
            no_wrap_idx = 0  # First half are no-wrap
            wrap_idx = self.n_envs // 2  # Second half are wrap
            
            training_state_1 = self.training_env.get_attr("session")[no_wrap_idx].get_state()
            training_state_1["wrap_around"] = False
            
            training_state_2 = self.training_env.get_attr("session")[wrap_idx].get_state()
            training_state_2["wrap_around"] = True
            
            eval_state_1 = self.eval_envs[0].unwrapped.session.get_state()
            eval_state_1["wrap_around"] = self.eval_envs[0].unwrapped.rules.wrap_around
            
            eval_state_2 = self.eval_envs[1].unwrapped.session.get_state()
            eval_state_2["wrap_around"] = self.eval_envs[1].unwrapped.rules.wrap_around
            
            demo_state = self.demo_env.unwrapped.session.get_state()
            demo_state["wrap_around"] = self.demo_env.unwrapped.rules.wrap_around
            
            # Update viz_data
            viz_data = {
                'training_state': [training_state_1, training_state_2],
                'eval_states': [eval_state_1, eval_state_2],
                'demo_state': demo_state,
                'weights_info': {
                    'action_probs': action_probs.tolist() if 'action_probs' in locals() else [0.25] * 4,
                    'value_estimate': float(value_estimate[0]) if 'value_estimate' in locals() else 0.0,
                    'action_labels': ['Up', 'Right', 'Down', 'Left']
                },
                'training_info': {
                    'total_timesteps': self.num_timesteps,
                    'initial_timesteps': self.initial_timesteps,  # Add initial timesteps
                    'target_timesteps': self.total_timesteps,
                    'episode_reward': float(self.current_reward),
                    'mean_reward': float(np.mean(self.episode_rewards[-100:])) if self.episode_rewards else 0.0,
                    'episode_length': int(self.current_length),
                    'mean_length': float(np.mean(self.episode_lengths[-100:])) if self.episode_lengths else 0.0,
                    'mean_eval_score': float(np.mean(self.eval_scores[-100:])) if self.eval_scores else 0.0,
                    'rewards_history': self.episode_rewards[-1000:],
                    'lengths_history': self.episode_lengths[-1000:],
                    'eval_scores_history': self.eval_scores[-1000:],
                    'high_score': self.high_score,
                    'recent_high_score': self.recent_high_score
                }
            }
            
            # Send update to visualization
            self.viz_queue.put(('update', viz_data))
            
            # Reset recent high score periodically
            if self.num_timesteps % (self.n_steps * self.n_envs) == 0:
                print(f"\nEvaluation High Score: {self.high_score}")
                print(f"Recent Evaluation High Score: {self.recent_high_score}")
                self.recent_high_score = 0
            
            return True
            
        except Exception as e:
            print(f"Error in visualization callback: {e}")
            import traceback
            traceback.print_exc()
            return True

def make_env(seed: int = 0, wrap_walls: bool = False) -> gym.Env:
    """Create a training environment.
    
    Args:
        seed: Random seed
        wrap_walls: Whether snake can wrap around walls
    """
    def _init() -> gym.Env:
        env = SnakeEnv(render_mode=None)
        env = Monitor(env)
        # Set wrap-around mode
        env.unwrapped.rules.wrap_around = wrap_walls
        env.reset(seed=seed)
        return env
    
    set_random_seed(seed)
    return _init

def train_model(
    total_timesteps: int,
    viz_queue: Queue,
    model_path: Optional[str] = None,
    cuda: bool = True,
    learning_rate: float = 3e-4,  # Standard PPO learning rate
    batch_size: int = 256,        # More reasonable batch size
    n_envs: int = 16,            # Balanced number of environments
    n_steps: int = 2048          # Standard PPO steps
) -> None:
    """
    Train the snake AI model.
    
    Args:
        total_timesteps: Number of timesteps to train for
        viz_queue: Queue for sending visualization updates
        model_path: Path to save/load the model
        cuda: Whether to use CUDA for training
        learning_rate: Learning rate for the optimizer
        batch_size: Batch size for training
        n_envs: Number of parallel environments
        n_steps: Number of steps per environment
    """
    try:
        # Set up environments - half with wrap, half without
        envs = []
        for i in range(n_envs):
            wrap_walls = i >= n_envs // 2  # Half the envs with wrap-around
            envs.append(make_env(i, wrap_walls))
        
        env = SubprocVecEnv(envs)
        
        # PPO hyperparameters - adjusted for more stable learning
        ppo_params = dict(
            learning_rate=1e-4,           # Slightly higher learning rate
            n_steps=1024,                # Shorter horizon for faster updates
            batch_size=128,              # Smaller batch size for more frequent updates
            n_epochs=4,                  # Balanced number of epochs
            gamma=0.98,                  # Slightly lower discount for more immediate rewards
            gae_lambda=0.9,              # Lower GAE lambda for more immediate advantage estimates
            clip_range=0.1,              # Smaller clip range for more conservative updates
            clip_range_vf=0.1,           # Also clip value function
            normalize_advantage=True,
            ent_coef=0.2,              # Moderate entropy for exploration
            vf_coef=0.8,                # Balanced value function coefficient
            max_grad_norm=0.3,          # More conservative gradient clipping
            target_kl=None,             
            verbose=1,
            device="cuda" if cuda and torch.cuda.is_available() else "cpu"
        )
        
        # Create or load model with updated architecture
        policy_kwargs = dict(
            features_extractor_class=CustomRecurrentCNN,  # Use the new recurrent extractor
            features_extractor_kwargs=dict(features_dim=512),  # You can adjust features_dim as needed
            net_arch=dict(
                pi=[512, 256, 128],  # Policy network layers
                vf=[512, 256, 128]   # Value network layers
            ),
            activation_fn=nn.ReLU,
            normalize_images=False
        )
        
        # Track initial timesteps for progress tracking
        initial_timesteps = 0
        
        if model_path and os.path.exists(model_path):
            print(f"Loading model from {model_path}")
            try:
                model = PPO.load(
                    model_path,
                    env=env,
                    custom_objects={"policy_kwargs": policy_kwargs},
                    **ppo_params
                )
                initial_timesteps = model.num_timesteps
                print(f"Loaded model with {initial_timesteps:,} timesteps of training")
            except Exception as e:
                print(f"Failed to load model with error: {e}")
                print("Creating new model instead")
                model = PPO("CnnPolicy", env, policy_kwargs=policy_kwargs, **ppo_params)
        else:
            print("Creating new model")
            model = PPO("CnnPolicy", env, policy_kwargs=policy_kwargs, **ppo_params)
        
        # Set up visualization callback with initial timesteps
        viz_callback = VisualizationCallback(
            viz_queue=viz_queue,
            total_timesteps=total_timesteps,
            eval_freq=1000,
            n_steps=n_steps,
            n_envs=n_envs,
            initial_timesteps=initial_timesteps  # Pass initial timesteps to callback
        )
        
        # Training loop with stop check
        should_stop = False
        timesteps_per_batch = n_steps * n_envs
        num_batches = total_timesteps // timesteps_per_batch
        
        for i in range(num_batches):
            if should_stop:
                break
                
            # Train for one batch
            model.learn(
                total_timesteps=timesteps_per_batch,
                callback=viz_callback,
                progress_bar=True,
                reset_num_timesteps=False
            )
            
            # Check for stop signal from visualization
            try:
                while True:  # Process all pending messages
                    msg_type, _ = viz_queue.get_nowait()
                    if msg_type == 'stop_training':
                        should_stop = True
                        break
            except Empty:
                pass
        
        # Save model if training completed or interrupted
        if model_path:
            print(f"Saving model to {model_path}")
            model.save(model_path)
        
    except Exception as e:
        print(f"Error in training process: {e}")
        import traceback
        traceback.print_exc()
    finally:
        # Clean up
        env.close()
        # Signal visualization to stop if we haven't already
        viz_queue.put(('stop', None))

def main():
    """Main training script."""
    parser = argparse.ArgumentParser(description="Train Snake AI")
    parser.add_argument("--config", action="store_true",
                       help="Show configuration UI before training")
    parser.add_argument("--model", type=str, default="models/snake_ai.zip",
                       help="Path to save/load model")
    parser.add_argument("--no-cuda", action="store_true",
                       help="Disable CUDA training")
    args = parser.parse_args()
    
    # Get training configuration
    if args.config:
        from src.ai.config_ui import ConfigUI
        config_ui = ConfigUI()
        config = config_ui.run()
        if config is None:  # User cancelled
            return
    else:
        # Load saved config or use defaults
        config_file = "training_config.json"
        if os.path.exists(config_file):
            with open(config_file, 'r') as f:
                config = json.load(f)
        else:
            config = {
                'timesteps': 2000000,
                'learning_rate': 0.0001,
                'batch_size': 512,
                'n_envs': 32,
                'n_steps': 512
            }
    
    # Create visualization queue
    viz_queue = Queue()
    
    # Start visualization in separate thread
    viz_thread = Thread(target=run_dashboard, args=(viz_queue,))
    viz_thread.start()
    
    # Start training with configuration
    train_model(
        total_timesteps=config['timesteps'],
        viz_queue=viz_queue,
        model_path=args.model,
        cuda=not args.no_cuda,
        learning_rate=config['learning_rate'],
        batch_size=config['batch_size'],
        n_envs=config['n_envs'],
        n_steps=config['n_steps']
    )
    
    # Wait for visualization to finish
    viz_thread.join()

if __name__ == "__main__":
    main() 