"""
Training Visualization Dashboard

This module provides a real-time visualization of the training process,
showing both the training metrics and evaluation games.
"""

import pygame
import numpy as np
from typing import Dict, Any, Optional, List, Tuple
import threading
from queue import Queue, Empty
import time

class TrainingDashboard:
    def __init__(self, width: int = 1920, height: int = 1080):
        """Initialize the training dashboard."""
        self.width = width
        self.height = height
        
        # Fonts
        self.title_font = None
        self.header_font = None
        self.text_font = None
        
        # Colors
        self.colors = {
            'background': (0, 0, 0),
            'text': (200, 200, 200),
            'highlight': (0, 255, 0),
            'grid': (40, 40, 40),
            'border': (0, 255, 0),
            'snake': (0, 255, 0),
            'food': (255, 0, 0),
            'graph_bg': (20, 20, 20),
            'graph_line': (0, 255, 0),
            'graph_grid': (40, 40, 40)
        }
        
        # Layout
        self.layout = {
            'margin': 20,
            'game_size': 400,
            'graph_height': 200,
            'metrics_width': 300
        }
        
        # Training metrics history
        self.metrics_history = {
            'rewards': [],
            'scores': [],
            'lengths': []
        }
        self.max_history = 1000
        
        # Current state
        self.current_state = None
        self.eval_state = None
        self.demo_state = None
        self.training_info = None
        self.model_info = None
        
        # Performance tracking
        self.frame_times = []
        self.max_frame_times = 60
        
        # Initialize the clock
        self.clock = None
        self.target_fps = 60
        
    def run(self, update_queue: Queue) -> None:
        """Run the dashboard, processing updates from the queue."""
        pygame.init()
        try:
            self.screen = pygame.display.set_mode((self.width, self.height))
            pygame.display.set_caption("Snake AI Training Dashboard")
            
            # Initialize fonts after pygame is initialized
            self.title_font = pygame.font.Font(None, 48)
            self.header_font = pygame.font.Font(None, 36)
            self.text_font = pygame.font.Font(None, 24)
            
            # Initialize clock
            self.clock = pygame.time.Clock()
            
            running = True
            while running:
                # Handle events
                for event in pygame.event.get():
                    if event.type == pygame.QUIT:
                        running = False
                    elif event.type == pygame.KEYDOWN:
                        if event.key == pygame.K_ESCAPE:
                            running = False
                
                # Process updates from queue
                try:
                    while True:  # Process all available updates
                        update_type, data = update_queue.get_nowait()
                        if update_type == 'update':
                            self.update(data)
                        elif update_type == 'stop':
                            running = False
                            break
                except Empty:
                    pass
                
                # Render dashboard
                self.render()
                self.clock.tick(self.target_fps)
        finally:
            pygame.quit()
    
    def update(self, data: Dict[str, Any]) -> None:
        """Update the dashboard with new data."""
        self.current_state = data  # Store entire data dict
        if 'training_info' in data:
            self.training_info = data['training_info']
            # Update metrics history
            for key in ['rewards', 'scores', 'lengths']:
                if key in self.training_info:
                    self.metrics_history[key].append(self.training_info[key])
                    if len(self.metrics_history[key]) > self.max_history:
                        self.metrics_history[key].pop(0)
    
    def draw_game_view(self, state: Dict[str, Any], position: Tuple[int, int], size: int, title: str) -> None:
        """Draw a game view at the specified position."""
        if not state:
            return
        
        x, y = position
        
        # Draw game area background
        game_rect = pygame.Rect(x, y, size, size)
        pygame.draw.rect(self.screen, self.colors['grid'], game_rect)
        pygame.draw.rect(self.screen, self.colors['border'], game_rect, 2)
        
        # Get time limit info directly from state
        steps_since_food = state.get('steps_since_food', 0)
        current_time_limit = state.get('current_time_limit', 100)  # Default to base limit if not provided
        time_left = max(0, current_time_limit - steps_since_food)
        
        # Calculate cell size based on grid dimensions
        cell_size = size // 30  # Assuming 30x30 grid
        
        # Draw snake
        if "snake_body" in state:
            for segment in state["snake_body"]:
                segment_rect = pygame.Rect(
                    x + segment[0] * cell_size,
                    y + segment[1] * cell_size,
                    cell_size, cell_size
                )
                pygame.draw.rect(self.screen, self.colors['snake'], segment_rect)
        
        # Draw food
        if "food_position" in state:
            food_rect = pygame.Rect(
                x + state["food_position"][0] * cell_size,
                y + state["food_position"][1] * cell_size,
                cell_size, cell_size
            )
            pygame.draw.rect(self.screen, self.colors['food'], food_rect)
        
        # Draw score and timer above game area
        score = state.get('score', 0)
        score_text = f"Score: {score}"
        timer_text = f"Time: {time_left}/{current_time_limit}"  # Show both current and max time
        combined_text = f"{score_text}  |  {timer_text}"
        score_surface = self.text_font.render(combined_text, True, self.colors['text'])
        score_rect = score_surface.get_rect(midtop=(x + size//2, y - 25))
        self.screen.blit(score_surface, score_rect)
        
        # Draw title below game area, including wrap mode for training view
        if title == "Training Episode" and 'wrap_around' in state:
            wrap_mode = "(Wrap)" if state['wrap_around'] else "(No Wrap)"
            title = f"{title} {wrap_mode}"
        title_text = title
        title_surface = self.header_font.render(title_text, True, self.colors['text'])
        title_rect = title_surface.get_rect(midtop=(x + size//2, y + size + 10))
        self.screen.blit(title_surface, title_rect)
    
    def draw_metrics(self, x: int, y: int, width: int) -> None:
        """Draw training metrics."""
        if not self.training_info:
            return
        
        font = pygame.font.Font(None, 24)
        y_offset = 0
        
        # Draw high scores
        high_score = self.training_info.get("high_score", 0)
        recent_high = self.training_info.get("recent_high_score", 0)
        
        metrics = [
            ("All-time High Score", high_score),
            ("Recent High Score", recent_high),
            ("", ""),  # Empty line for spacing
            ("Training Metrics", ""),
            ("Total Steps", self.training_info.get('total_timesteps', 0)),
            ("Episode Reward", f"{self.training_info.get('episode_reward', 0):.2f}"),
            ("Mean Reward", f"{self.training_info.get('mean_reward', 0):.2f}"),
            ("Episode Length", self.training_info.get('episode_length', 0)),
            ("Mean Length", f"{self.training_info.get('mean_length', 0):.2f}"),
            ("FPS", f"{len(self.frame_times) / sum(self.frame_times):.1f}" if self.frame_times else "0")
        ]
        
        for i, (label, value) in enumerate(metrics):
            text = f"{label}: {value}" if label else ""
            surface = self.text_font.render(text, True, self.colors['text'])
            self.screen.blit(surface, (x, y + i * 25))
    
    def draw_graph(self, data: List[float], rect: pygame.Rect, title: str,
                  color: Tuple[int, int, int], min_val: float, max_val: float,
                  smoothing_window: int = 10) -> None:
        """Draw a line graph of the given data.

        Args:
            data: List of data points.
            rect: Rectangle area for drawing the graph.
            title: Title for the graph.
            color: Color for the graph line.
            min_val: Minimum value for scaling the y-axis.
            max_val: Maximum value for scaling the y-axis.
            smoothing_window: Optional smoothing window size for drawing
                long term trends. Increase this value for a smoother, longer-term trend.
                Default is 10.
        """
        if not data:
            return

        # Create a separate surface for the graph
        graph_surface = pygame.Surface((rect.width, rect.height))
        graph_surface.fill(self.colors['graph_bg'])

        # Draw grid lines on the graph surface
        num_lines = 5
        for i in range(num_lines):
            y_line = i * rect.height / (num_lines - 1)
            pygame.draw.line(graph_surface, self.colors['graph_grid'],
                             (0, y_line), (rect.width, y_line))

        # Add padding to the min/max values for better visualization
        value_range = max_val - min_val
        if value_range == 0:
            value_range = 1.0
        padding = value_range * 0.1
        min_val_adjusted = min_val - padding
        max_val_adjusted = max_val + padding

        # Smooth the data using the provided smoothing_window parameter
        window_size = smoothing_window if len(data) >= smoothing_window else len(data)
        if window_size > 1:
            smoothed_data = np.convolve(data, np.ones(window_size) / window_size, mode='valid')
        else:
            smoothed_data = data

        points = []
        for i, value in enumerate(smoothed_data):
            x_point = i * rect.width / len(smoothed_data)
            y_point = rect.height - ((value - min_val_adjusted) * rect.height /
                                     (max_val_adjusted - min_val_adjusted))
            points.append((x_point, y_point))

        if len(points) > 1:
            pygame.draw.lines(graph_surface, color, False, points, 2)

        # Draw border on the graph surface
        pygame.draw.rect(graph_surface, self.colors['border'],
                         pygame.Rect(0, 0, rect.width, rect.height), 1)

        # Blit the graph surface onto the main screen
        self.screen.blit(graph_surface, rect)

        # Draw title and current value
        current_value = data[-1] if data else 0
        title_text = f"{title} (Current: {current_value:.2f})"
        title_surface = self.text_font.render(title_text, True, self.colors['text'])
        title_rect = title_surface.get_rect(midtop=(rect.centerx, rect.top - 20))
        self.screen.blit(title_surface, title_rect)

        # Draw value labels to the right of the graph
        label_spacing = rect.height / (num_lines - 1)
        label_x = rect.right + 10
        for i in range(num_lines):
            value = max_val_adjusted - (i * (max_val_adjusted - min_val_adjusted) / (num_lines - 1))
            label_text = f"{value:.1f}"
            label_surface = self.text_font.render(label_text, True, self.colors['text'])
            label_rect = label_surface.get_rect(
                left=label_x,
                centery=rect.top + i * label_spacing
            )
            self.screen.blit(label_surface, label_rect)
    
    def draw_weights_visualization(self, x: int, y: int, width: int, height: int, weights_info: Dict) -> None:
        """Draw the neural network weights visualization."""
        if not weights_info:
            return
            
        # Get data
        action_probs = weights_info.get('action_probs', [0, 0, 0, 0])
        value_estimate = weights_info.get('value_estimate', 0.0)
        action_labels = weights_info.get('action_labels', ['Up', 'Right', 'Down', 'Left'])
        
        # Draw background
        rect = pygame.Rect(x, y, width, height)
        pygame.draw.rect(self.screen, self.colors['graph_bg'], rect)
        pygame.draw.rect(self.screen, self.colors['border'], rect, 1)
        
        # Draw title
        title = "Action Probabilities"
        title_surface = self.header_font.render(title, True, self.colors['text'])
        title_rect = title_surface.get_rect(midtop=(x + width//2, y + 5))
        self.screen.blit(title_surface, title_rect)
        
        # Draw action probability bars
        bar_height = 20
        bar_spacing = 30
        bar_start_y = y + 50
        max_bar_width = width - 120  # Leave space for labels
        
        for i, (prob, label) in enumerate(zip(action_probs, action_labels)):
            # Draw label
            label_surface = self.text_font.render(label, True, self.colors['text'])
            self.screen.blit(label_surface, (x + 10, bar_start_y + i * bar_spacing))
            
            # Draw bar background
            bar_bg_rect = pygame.Rect(x + 70, bar_start_y + i * bar_spacing, max_bar_width, bar_height)
            pygame.draw.rect(self.screen, self.colors['grid'], bar_bg_rect)
            
            # Draw probability bar
            bar_width = int(prob * max_bar_width)
            bar_rect = pygame.Rect(x + 70, bar_start_y + i * bar_spacing, bar_width, bar_height)
            pygame.draw.rect(self.screen, self.colors['highlight'], bar_rect)
            
            # Draw probability value
            prob_text = f"{prob:.2f}"
            prob_surface = self.text_font.render(prob_text, True, self.colors['text'])
            self.screen.blit(prob_surface, (x + 80 + max_bar_width, bar_start_y + i * bar_spacing))
        
        # Draw value estimate
        value_text = f"Value Estimate: {value_estimate:.2f}"
        value_surface = self.text_font.render(value_text, True, self.colors['text'])
        value_rect = value_surface.get_rect(midbottom=(x + width//2, y + height - 10))
        self.screen.blit(value_surface, value_rect)

    def draw_progress_bar(self, x: int, y: int, width: int, height: int, progress: float, text: str) -> None:
        """Draw a progress bar with text.
        
        Args:
            x, y: Position of the progress bar
            width, height: Dimensions of the progress bar
            progress: Progress value between 0 and 1
            text: Text to display above the progress bar
        """
        # Draw background
        bg_rect = pygame.Rect(x, y, width, height)
        pygame.draw.rect(self.screen, self.colors['graph_bg'], bg_rect)
        pygame.draw.rect(self.screen, self.colors['border'], bg_rect, 1)
        
        # Draw progress
        if progress > 0:
            progress_width = int(width * progress)
            progress_rect = pygame.Rect(x, y, progress_width, height)
            pygame.draw.rect(self.screen, self.colors['highlight'], progress_rect)
        
        # Draw text
        text_surface = self.text_font.render(text, True, self.colors['text'])
        text_rect = text_surface.get_rect(bottomleft=(x, y - 5))
        self.screen.blit(text_surface, text_rect)

    def render(self) -> None:
        """Render the dashboard."""
        if not self.current_state:
            return

        self.screen.fill(self.colors['background'])

        # Draw game views
        if 'training_state' in self.current_state:
            # Draw both training games
            self.draw_game_view(
                self.current_state['training_state'][0],  # No wrap training
                (50, 50),
                250,
                "Training (No Wrap)"
            )
            self.draw_game_view(
                self.current_state['training_state'][1],  # Wrap training
                (350, 50),
                250,
                "Training (Wrap)"
            )

        if 'eval_states' in self.current_state:
            self.draw_game_view(
                self.current_state['eval_states'][0],
                (650, 50),
                250,
                "Evaluation (No Wrap)"
            )
            self.draw_game_view(
                self.current_state['eval_states'][1],
                (950, 50),
                250,
                "Evaluation (Wrap)"
            )

        if 'demo_state' in self.current_state:
            self.draw_game_view(
                self.current_state['demo_state'],
                (1250, 50),
                250,
                "Demo Game"
            )

        # Draw metrics
        if self.training_info:
            metrics_data = [
                (self.training_info['rewards_history'], "Episode Rewards"),
                (self.training_info['lengths_history'], "Episode Lengths"),
                (self.training_info['eval_scores_history'], "Evaluation Scores")
            ]

            for i, (data, label) in enumerate(metrics_data):
                if data:  # Only draw if we have data
                    rect = pygame.Rect(50, 400 + i * 200, 500, 150)
                    self.draw_graph(
                        data,
                        rect,
                        label,
                        self.colors['graph_line'],
                        min(data) if data else 0,
                        max(data) if data else 1
                    )

            # Draw current metrics text
            metrics_text = [
                f"Total Steps: {self.training_info.get('total_timesteps', 0):,}",
                f"Mean Reward: {self.training_info.get('mean_reward', 0):.2f}",
                f"Mean Length: {self.training_info.get('mean_length', 0):.1f}",
                f"Mean Eval Score: {self.training_info.get('mean_eval_score', 0):.2f}",
                f"High Score: {self.training_info.get('high_score', 0)}",
                f"Recent High Score: {self.training_info.get('recent_high_score', 0)}"
            ]

            # Add vertical spacer between metrics sections
            metrics_text.insert(3, "")  # Insert empty string as spacer after first 3 metrics

            font = pygame.font.Font(None, 36)
            for i, text in enumerate(metrics_text):
                surface = font.render(text, True, self.colors['text'])
                self.screen.blit(surface, (600, 400 + i * 30))

        # Draw action probabilities if available
        if 'weights_info' in self.current_state:
            self.draw_weights_visualization(
                1100, 400,
                300, 300,
                self.current_state['weights_info']
            )
            
            # Draw trend indicator graph
            if self.training_info:
                # Calculate trend indicators
                window = 100  # Use last 100 episodes for trends
                rewards = self.training_info['rewards_history'][-window:]
                lengths = self.training_info['lengths_history'][-window:]
                eval_scores = self.training_info['eval_scores_history'][-window:]
                
                if rewards and lengths and eval_scores:
                    # Normalize each metric to 0-1 range for fair combination
                    norm_rewards = [(r - min(rewards)) / (max(rewards) - min(rewards) + 1e-8) for r in rewards]
                    norm_lengths = [(l - min(lengths)) / (max(lengths) - min(lengths) + 1e-8) for l in lengths]
                    norm_scores = [(s - min(eval_scores)) / (max(eval_scores) - min(eval_scores) + 1e-8) for s in eval_scores]
                    
                    # Combine metrics with weights (rewards: 0.4, lengths: 0.3, eval_scores: 0.3)
                    trend_data = [0.4 * r + 0.3 * l + 0.3 * s 
                                for r, l, s in zip(norm_rewards, norm_lengths, norm_scores)]
                                        # Add trend direction indicator

                    trending = 0 # -1, 0, 1
                
                    if len(trend_data) >= 2:
                        recent_trend = trend_data[-1] - trend_data[0]
                        trending = 1 if recent_trend > 0.1 else -1 if recent_trend < -0.1 else 0
                        trend_text = "↑ Improving" if trending == 1 else "↓ Declining" if trending == -1 else "→ Stable"
                        trend_color = (0, 255, 0) if trending == 1 else (255, 0, 0) if trending == -1 else (200, 200, 0)
                        trend_surface = self.text_font.render(trend_text, True, trend_color)
                        trend_rect = trend_surface.get_rect(midtop=(1250, 860))
                        self.screen.blit(trend_surface, trend_rect)

                    graph_color = (0, 255, 0) if trending == 1 else (255, 0, 0) if trending == -1 else (200, 200, 0)

                    # Draw trend graph 50 pixels below the action probabilities graph
                    trend_rect = pygame.Rect(1100, 750, 300, 150)
                    self.draw_graph(
                        trend_data,
                        trend_rect,
                        "Learning Trend",
                        graph_color,  # color based on trend
                        0,  # Min value (normalized)
                        1,  # Max value (normalized)
                        smoothing_window=10  # Use a larger smoothing window for longer-term trends
                    )

        # Draw training progress bar at the bottom of the screen
        if self.training_info:
            # Get current session steps (from start of this training run)
            current_session_steps = self.training_info.get('total_timesteps', 0) - self.training_info.get('initial_timesteps', 0)
            target_timesteps = self.training_info.get('target_timesteps', 1000000)
            progress = min(1.0, current_session_steps / target_timesteps)
            
            # Show both current session progress and total model training
            total_model_steps = self.training_info.get('total_timesteps', 0)
            progress_text = (
                f"Current Session: {current_session_steps:,}/{target_timesteps:,} steps ({progress*100:.1f}%) | "
                f"Total Model Training: {total_model_steps:,} steps"
            )
            self.draw_progress_bar(50, self.height - 40, self.width - 100, 20, progress, progress_text)

        pygame.display.flip()
    
    def handle_events(self):
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                self.running = False
            elif event.type == pygame.KEYDOWN:
                if event.key == pygame.K_ESCAPE:
                    self.running = False

def run_dashboard(update_queue: Queue) -> None:
    """Run the training dashboard in a separate thread."""
    dashboard = TrainingDashboard()
    try:
        dashboard.run(update_queue)
    finally:
        # Send stop signal back to training process
        update_queue.put(('stop_training', None))
        pygame.quit()