"""
Game Settings

This module contains the game settings and rules that can be configured.
"""

class GameRules:
    """Game rules that can be configured through the settings menu."""
    
    def __init__(self):
        self.wrap_around = True  # Whether snake wraps around screen edges
        self.speed_increase = True  # Whether snake speeds up as it grows
        self.min_move_cooldown = 50  # Minimum movement delay in milliseconds
        self.initial_move_cooldown = 100  # Initial movement delay
        self.starting_length = 3  # Starting length of the snake

    def update_rule(self, rule_name, value):
        """
        Update a game rule with a new value.
        
        Args:
            rule_name (str): Name of the rule to update
            value: New value for the rule
        
        Returns:
            bool: True if update was successful, False otherwise
        """
        if hasattr(self, rule_name):
            # Define validation rules for specific attributes
            validators = {
                'starting_length': lambda x: 1 <= x <= 10,
                'wrap_around': lambda x: isinstance(x, bool),
                'speed_increase': lambda x: isinstance(x, bool),
                'min_move_cooldown': lambda x: x > 0,
                'initial_move_cooldown': lambda x: x > 0
            }
            
            # Validate the value if there's a validator for this rule
            if rule_name in validators and not validators[rule_name](value):
                return False
                
            setattr(self, rule_name, value)
            return True
        return False

class GameSettings:
    """Global game settings."""
    
    def __init__(self):
        self.debug_mode = False  # Whether to show debug information
        self.show_grid = self.debug_mode  # Whether to show the grid lines
        self.rules = GameRules()  # Game rules instance 