"""
Food Module

This module provides the Food class that represents the food item in the game.
"""

import pygame
import random
from typing import Tuple, List
from src.ui import GameArea
from src.config.colors import *

class Food:
    def __init__(self, color: Tuple[int, int, int] = NEON_RED):
        """
        Initialize food with a color.
        
        Args:
            color: RGB color tuple for the food
        """
        self.color = color
        self.position = (0, 0)  # Will be set by spawn_at_position()

    def spawn_at_position(self, available_positions: List[Tuple[int, int]]) -> None:
        """
        Spawn food at a random position from the available positions.
        
        Args:
            available_positions: List of valid grid positions where food can spawn
        """
        if not available_positions:
            # No valid positions (snake fills screen) - game should be won
            return
        
        # Choose random position from valid positions
        self.position = random.choice(available_positions)

    def draw(self, screen: pygame.Surface, game_area: GameArea) -> None:
        """
        Draw the food on the screen with a glowing effect
        
        Args:
            screen: Pygame surface to draw on
            game_area: GameArea instance for coordinate conversion
        """
        screen_x, screen_y = game_area.get_screen_pos(self.position[0], self.position[1])
        
        # Draw outer glow
        glow_size = game_area.grid_size + 4
        glow_surface = pygame.Surface((glow_size, glow_size), pygame.SRCALPHA)
        
        # Create radial gradient for glow
        for i in range(3):
            alpha = 100 - (i * 30)  # Fade out alpha
            size = glow_size - (i * 2)
            pos = (glow_size - size) // 2
            pygame.draw.circle(glow_surface, (*NEON_RED[:3], alpha), (glow_size//2, glow_size//2), size//2)
        
        # Draw glow
        screen.blit(glow_surface, (screen_x - 2, screen_y - 2))
        
        # Draw main food body
        food_rect = pygame.Rect(
            screen_x + 2,
            screen_y + 2,
            game_area.grid_size - 4,
            game_area.grid_size - 4
        )
        pygame.draw.rect(screen, NEON_RED, food_rect, border_radius=4)
        
        # Draw highlight
        highlight_rect = pygame.Rect(
            screen_x + 4,
            screen_y + 4,
            game_area.grid_size - 8,
            game_area.grid_size - 8
        )
        pygame.draw.rect(screen, (*NEON_RED, 200), highlight_rect, border_radius=3)

    def check_collision(self, position: Tuple[int, int]) -> bool:
        """Check if the given position collides with the food"""
        return self.position == position 