"""
Game Session Manager

This module provides a reusable game session class that encapsulates the core gameplay mechanics.
It can be used by both the main game and the training environment to ensure consistent behavior.
"""

import pygame
from typing import Tuple, Dict, Optional
from src.core import Snake, Direction, Food
from src.ui import GameArea
from src.config import (
    WINDOW_WIDTH,
    WINDOW_HEIGHT,
    FPS
)
from src.config.colors import *
from src.config.settings import GameRules

class GameSession:
    """Manages a single game session with consistent rules and boundaries."""
    
    def __init__(self, grid_width: int, grid_height: int, rules : GameRules, window_width: int = WINDOW_WIDTH, window_height: int = WINDOW_HEIGHT):
        """
        Initialize a new game session.
        
        Args:
            grid_width: Number of grid cells horizontally
            grid_height: Number of grid cells vertically
            rules: GameRules instance containing game settings
            window_width: Width of the game window
            window_height: Height of the game window
        """
        self.grid_width = grid_width
        self.grid_height = grid_height
        self.rules = rules
        self.window_width = window_width
        self.window_height = window_height
        
        # Game objects
        self.snake = None
        self.food = None
        
        # Game state
        self.score = 0
        self.is_game_over = False
        self.move_cooldown = rules.initial_move_cooldown
        self.last_move_time = 0
        
        self.reset()
    
    def reset(self) -> Dict:
        """
        Reset the game session to its initial state, including the game area.
        
        Returns:
            Dict containing the initial game state
        """
        # Initialize game area with window dimensions
        self.game_area = GameArea(
            window_width=self.window_width,
            window_height=self.window_height,
            grid_width=self.grid_width,
            grid_height=self.grid_height
        )

        # Initialize snake in the middle of the grid
        start_x = self.grid_width // 2
        start_y = self.grid_height // 2
        self.snake = Snake(
            start_pos=(start_x, start_y),
            grid_width=self.grid_width,
            grid_height=self.grid_height
        )
        
        # Initialize food
        self.food = Food()
        self._spawn_food()
        
        # Reset game state
        self.score = 0
        self.is_game_over = False
        self.move_cooldown = self.rules.initial_move_cooldown
        self.last_move_time = 0
        
        return self.get_state()
    
    def _spawn_food(self) -> None:
        """Spawn food in a random empty grid cell."""
        # Get all occupied positions
        occupied = set(self.snake.body)
        
        # Get all possible positions
        all_positions = [(x, y) for x in range(self.grid_width) 
                        for y in range(self.grid_height)]
        
        # Filter out occupied positions
        available = [pos for pos in all_positions if pos not in occupied]
        
        if available:
            self.food.spawn_at_position(available)
    
    def step(self, action: Optional[Direction] = None, current_time: int = 0) -> Tuple[Dict, float, bool]:
        """
        Advance the game state by one step.
        
        Args:
            action: Optional direction to change to
            current_time: Current game time in milliseconds
        
        Returns:
            Tuple of (game_state, reward, done)
        """
        if self.is_game_over:
            return self.get_state(), 0, True
        
        # Apply action if provided
        if action is not None:
            self.snake.change_direction(action)
        
        # Update snake movement animation
        self.snake.update_movement(1.0 / FPS) #
        
        # Check if enough time has passed for next move
        if current_time - self.last_move_time < self.move_cooldown:
            return self.get_state(), 0, False
        
        self.last_move_time = current_time
        
        # Process any pending direction changes first
        if self.snake.input_buffer:
            # Try first direction
            next_direction = self.snake.input_buffer[0]
            if self.snake.is_direction_valid(next_direction):
                self.snake.direction = next_direction
                self.snake.input_buffer.pop(0)
            # If invalid and we have another input, try that one first
            elif len(self.snake.input_buffer) > 1:
                alt_direction = self.snake.input_buffer[1]
                if self.snake.is_direction_valid(alt_direction):
                    self.snake.direction = alt_direction
                    self.snake.input_buffer.pop(1)  # Remove the successful second input
                # Remove the first input regardless as it's had its chance
                self.snake.input_buffer.pop(0)
            else:
                # Single invalid input, just remove it
                self.snake.input_buffer.pop(0)
        
        # Now get movement vector with updated direction
        dx, dy = self.snake.direction.to_vector()
        new_head = (self.snake.body[0][0] + dx, self.snake.body[0][1] + dy)
        
        # Check for immediate wall collision (before moving)
        if not self.rules.wrap_around:
            if (new_head[0] < 0 or new_head[0] >= self.grid_width or 
                new_head[1] < 0 or new_head[1] >= self.grid_height):
                self.is_game_over = True
                return self.get_state(), -1, True

        # check for collision with snake body
        if new_head in self.snake.body[1:]:
            self.is_game_over = True
            return self.get_state(), -1, True
        
        # Move snake
        head_x, head_y = self.snake.move()
        
        # Handle wrap-around if enabled
        if self.rules.wrap_around:
            head_x = head_x % self.grid_width
            head_y = head_y % self.grid_height
            self.snake.body[0] = (head_x, head_y)
            self.snake.target_positions[0] = (head_x, head_y)  # Update target position for smooth movement
        
        # Check self collision
        if (head_x, head_y) in self.snake.body[1:]:
            self.is_game_over = True
            return self.get_state(), -1, True
        
        # Check food collision
        if (head_x, head_y) == self.food.position:
            self.score += 1
            self.snake.grow()
            self._spawn_food()
            
            # Increase speed if enabled
            if self.rules.speed_increase:
                self.move_cooldown = max(
                    self.rules.min_move_cooldown,
                    self.move_cooldown - 2
                )
            return self.get_state(), 1, False
        
        return self.get_state(), 0, False
    
    def get_state(self) -> Dict:
        """
        Get the current game state.
        
        Returns:
            Dict containing the game state
        """
        return {
            "grid_width": self.grid_width,
            "grid_height": self.grid_height,
            "snake_head": self.snake.body[0],
            "snake_body": self.snake.body,
            "snake_direction": self.snake.direction,
            "food_position": self.food.position,
            "score": self.score,
            "game_over": self.is_game_over,
            "move_cooldown": self.move_cooldown
        }
    
    def render(self, screen: pygame.Surface) -> None:
        """
        Render the game state to the screen.
        
        Args:
            screen: Pygame surface to render to
        """
        # Draw game area first
        self.game_area.draw(screen)
        
        # Draw game objects
        self.snake.draw(screen, self.game_area)
        self.food.draw(screen, self.game_area)
        
        # Draw grid (optional, for debugging)
        if hasattr(self, 'show_grid') and self.show_grid:
            for x in range(self.grid_width + 1):
                start_pos = (self.game_area.x + x * self.game_area.grid_size, self.game_area.y)
                end_pos = (self.game_area.x + x * self.game_area.grid_size, self.game_area.y + self.game_area.height)
                pygame.draw.line(screen, GRID_COLOR, start_pos, end_pos, 1)
            
            for y in range(self.grid_height + 1):
                start_pos = (self.game_area.x, self.game_area.y + y * self.game_area.grid_size)
                end_pos = (self.game_area.x + self.game_area.width, self.game_area.y + y * self.game_area.grid_size)
                pygame.draw.line(screen, GRID_COLOR, start_pos, end_pos, 1) 