"""
Snake Module

This module provides the Snake class and Direction enum for snake movement and rendering.
"""

import pygame
from enum import Enum, auto
from typing import Tuple, List
import math
from src.config.colors import *
import pygame.gfxdraw

class Direction(Enum):
    UP = auto()
    DOWN = auto()
    LEFT = auto()
    RIGHT = auto()
    
    def to_vector(self) -> Tuple[int, int]:
        # Convert direction to a movement vector
        if self == Direction.UP:
            return (0, -1)
        elif self == Direction.DOWN:
            return (0, 1)
        elif self == Direction.LEFT:
            return (-1, 0)
        else:  # Direction.RIGHT
            return (1, 0)

class Snake:
    def __init__(self, start_pos: Tuple[int, int], grid_width: int, grid_height: int, length: int = 3):
        """
        Initialize snake at the given grid position.
        
        Args:
            start_pos: Starting position in grid coordinates (x, y)
            grid_width: Width of the game grid
            grid_height: Height of the game grid
            length: Initial length of the snake
        """
        self.direction = Direction.RIGHT
        self.body = [start_pos]  # Head is at index 0
        self.growing = False
        self.grid_width = grid_width
        self.grid_height = grid_height
        
        # FIX: Start from 1 to avoid duplicate head segment.
        for i in range(1, length):
            self.body.append((start_pos[0] - i, start_pos[1]))

        # Initialize visual positions for each body segment
        self.visual_positions = [(float(pos[0]), float(pos[1])) for pos in self.body]
        self.move_progress = 1.0  # Progress of current move (0.0 to 1.0)
        self.move_speed = 0.2    # Movement speed
        self.target_positions = list(self.body)  # Target grid positions
        
        # Input buffer for direction changes
        self.input_buffer = []  # Store up to 2 pending direction changes
        self.max_buffer_size = 2

        # Create base segment surface with gradient
        self.segment_size = 32  # Base size for segment surface
        self.base_surface = pygame.Surface((self.segment_size, self.segment_size), pygame.SRCALPHA)
        
        # Create a radial gradient for the base segment
        center = self.segment_size // 2
        for radius in range(center, -1, -1):
            # Calculate color intensity based on distance from center
            intensity = 1.0 - (radius / center) ** 0.8  # Adjust power for gradient shape
            color = (
                int(SNAKE_GRADIENT_START[0] * intensity),
                int(SNAKE_GRADIENT_START[1] * intensity),
                int(SNAKE_GRADIENT_START[2] * intensity),
                255
            )
            pygame.draw.circle(self.base_surface, color, (center, center), radius)

    def move(self) -> Tuple[int, int]:
        """
        Move the snake one grid cell in current direction.
        
        Returns:
            New head position (x, y)
        """
        # Get new head position
        head = self.body[0]
        new_head = self._get_new_head_position(head)
        
        # Update grid positions
        self.body.insert(0, new_head)
        if not self.growing:
            self.body.pop()
        else:
            self.growing = False
            
        # Update visual positions
        self.target_positions = list(self.body)  # Copy current grid positions as targets
        if len(self.visual_positions) < len(self.target_positions):
            self.visual_positions.append(self.visual_positions[-1])
        
        # Reset move progress for smooth animation
        self.move_progress = 0.0
        
        return new_head

    def update_movement(self, dt: float):
        """Update visual position interpolation with consistent timing."""
        if self.move_progress >= 1.0:
            return

        # Use cubic easing for smoother acceleration/deceleration
        self.move_progress = min(1.0, self.move_progress + self.move_speed)
        t = self.move_progress
        overall_progress = t * t * (3 - 2 * t)
        
        # Update each segment with fixed delay windows
        segment_count = len(self.visual_positions)
        for i in range(segment_count):
            if i >= len(self.target_positions):
                continue
            
            current = self.visual_positions[i]
            target = self.target_positions[i]
            
            dx = target[0] - current[0]
            dy = target[1] - current[1]
            
            # Check if this segment needs to wrap
            wrap_x = abs(dx) > 2
            wrap_y = abs(dy) > 2
            
            if wrap_x or wrap_y:
                self.visual_positions[i] = target
            else:
                # Use fixed delay windows instead of compounding delays
                delay_window = 0.2  # Total delay spread across all segments
                segment_delay = (i / segment_count) * delay_window
                
                # Calculate progress for this segment
                segment_progress = max(0.0, (overall_progress - segment_delay) / (1.0 - segment_delay))
                segment_progress = min(1.0, segment_progress)
                
                # Apply smoothed movement
                new_x = current[0] + dx * segment_progress
                new_y = current[1] + dy * segment_progress
                self.visual_positions[i] = (new_x, new_y)

    def _get_new_head_position(self, head: Tuple[int, int]) -> Tuple[int, int]:
        """Calculate new head position based on current direction"""
        x, y = head
        dx, dy = self.direction.to_vector()
        return (x + dx, y + dy)

    def grow(self):
        """Mark the snake to grow on next move"""
        self.growing = True

    def check_collision(self, width: int, height: int, wrap_around: bool = False) -> bool:
        """
        Check if snake has collided with walls or itself.
        
        Args:
            width: Game area width
            height: Game area height
            wrap_around: If True, snake wraps around screen edges instead of colliding
        """
        head = self.body[0]
        
        if not wrap_around:
            # Check wall collision
            if (head[0] < 0 or head[0] >= width or 
                head[1] < 0 or head[1] >= height):
                return True
        
        # Check self collision (skip head)
        if head in self.body[1:]:
            return True
        
        return False

    def wrap_position(self, width: int, height: int):
        """Wrap snake's head position around screen edges"""
        head_x, head_y = self.body[0]
        wrapped_head = (
            head_x % width,
            head_y % height
        )
        self.body[0] = wrapped_head

    def is_direction_valid(self, new_direction: Direction) -> bool:
        """
        Try to change direction, ensuring no 180-degree turns.
        
        Args:
            new_direction: The direction to change to
            
        Returns:
            True if direction was changed, False otherwise
        """
        opposite_directions = {
            Direction.UP: Direction.DOWN,
            Direction.DOWN: Direction.UP, 
            Direction.LEFT: Direction.RIGHT,
            Direction.RIGHT: Direction.LEFT
        }
        
        if opposite_directions[new_direction] == self.direction or new_direction == self.direction:
            return False
            
        self.direction = new_direction
        return True

    def change_direction(self, new_direction: Direction):
        """
        Buffer a direction change to be applied on next move.
        
        Args:
            new_direction: The direction to change to
        """
        if self.input_buffer and self.input_buffer[-1] == new_direction:
            return
        
        if len(self.input_buffer) < self.max_buffer_size:
            self.input_buffer.append(new_direction)

    def draw(self, screen: pygame.Surface, game_area):
        """Draw snake with clean, connected design and cute tongue."""
        if len(self.visual_positions) < 2:
            return

        base_size = game_area.grid_size
        segment_size = base_size - 4
        
        # Draw snake body segments
        for i in range(len(self.visual_positions) - 1):
            pos1 = self.visual_positions[i]
            pos2 = self.visual_positions[i + 1]
            
            screen_positions = self._get_wrapped_segment_positions(pos1, pos2, game_area)
            if not screen_positions:
                continue
            
            for sp1, sp2 in screen_positions:
                # Calculate segment properties
                progress = max(0.4, 1.0 - (i / len(self.visual_positions)))
                
                color = (
                    int(SNAKE_GRADIENT_START[0] * progress),
                    int(SNAKE_GRADIENT_START[1] * progress),
                    int(SNAKE_GRADIENT_START[2] * progress)
                )
                
                # Draw connecting segments with extra thickness
                center1 = (sp1[0] + base_size//2, sp1[1] + base_size//2)
                center2 = (sp2[0] + base_size//2, sp2[1] + base_size//2)
                
                # Draw a thicker base line first
                pygame.draw.line(screen, color, center1, center2, segment_size + 4)
                
                # Draw larger circles at joints to ensure corner coverage
                pygame.draw.circle(screen, color, center1, (segment_size + 2)//2)
                pygame.draw.circle(screen, color, center2, (segment_size + 2)//2)
                
                # If this is a corner (check by looking at next segment)
                if i < len(self.visual_positions) - 2:
                    pos3 = self.visual_positions[i + 2]
                    if pos2[0] != pos1[0] and pos3[1] != pos2[1] or \
                       pos2[1] != pos1[1] and pos3[0] != pos2[0]:
                        # Draw extra circle at the corner to ensure coverage
                        pygame.draw.circle(screen, color, center2, (segment_size + 6)//2)
        
        # Draw head
        head_pos = self.visual_positions[0]
        screen_pos = game_area.get_screen_pos(head_pos[0], head_pos[1])
        center = (screen_pos[0] + base_size//2, screen_pos[1] + base_size//2)
        
        # Draw head circle
        pygame.draw.circle(screen, SNAKE_GRADIENT_START, center, segment_size//2)
        
        # Calculate direction for eyes and tongue
        next_pos = self.visual_positions[1] if len(self.visual_positions) > 1 else head_pos
        dx = head_pos[0] - next_pos[0]  # Reversed the direction calculation
        dy = head_pos[1] - next_pos[1]  # Reversed the direction calculation
        angle = math.atan2(dy, dx)
        
        # Draw eyes
        eye_offset = segment_size//4
        eye_size = 2
        eye_angle = angle + math.pi/2
        
        for side in [-1, 1]:
            eye_x = center[0] + math.cos(eye_angle) * eye_offset * side
            eye_y = center[1] + math.sin(eye_angle) * eye_offset
            pygame.draw.circle(screen, (0, 0, 0), (int(eye_x), int(eye_y)), eye_size)
        
        # Draw flickering tongue with more pronounced animation
        time = pygame.time.get_ticks()
        tongue_flick = math.sin(time * 0.01) * 0.8 + 0.2  # Slower, more pronounced flicking
        
        # Tongue base position (at front of head)
        tongue_base_x = center[0] + math.cos(angle) * (segment_size//2)
        tongue_base_y = center[1] + math.sin(angle) * (segment_size//2)
        
        # Tongue length varies with flicking animation
        tongue_length = (3 + tongue_flick * 6)  # Length varies between 3 and 9 pixels
        fork_length = 4  # Slightly longer fork
        fork_angle = math.pi/3  # Wider fork angle
        
        # Calculate tongue tip
        tongue_tip_x = tongue_base_x + math.cos(angle) * tongue_length
        tongue_tip_y = tongue_base_y + math.sin(angle) * tongue_length
        
        # Calculate fork tips
        left_fork_x = tongue_tip_x + math.cos(angle + fork_angle) * fork_length
        left_fork_y = tongue_tip_y + math.sin(angle + fork_angle) * fork_length
        right_fork_x = tongue_tip_x + math.cos(angle - fork_angle) * fork_length
        right_fork_y = tongue_tip_y + math.sin(angle - fork_angle) * fork_length
        
        # Draw tongue with brighter red
        tongue_color = (255, 0, 0)  # Bright red
        
        # Draw main tongue line
        pygame.draw.line(screen, tongue_color, 
                        (tongue_base_x, tongue_base_y), 
                        (tongue_tip_x, tongue_tip_y), 2)
        
        # Draw forked tips
        pygame.draw.line(screen, tongue_color,
                        (tongue_tip_x, tongue_tip_y),
                        (left_fork_x, left_fork_y), 2)
        pygame.draw.line(screen, tongue_color,
                        (tongue_tip_x, tongue_tip_y),
                        (right_fork_x, right_fork_y), 2)

    def _get_wrapped_segment_positions(self, pos1, pos2, game_area):
        """
        Get screen positions for segment rendering, handling wrap-around.
        Only returns positions for segments that are close enough to interpolate.
        """
        grid_w = self.grid_width
        grid_h = self.grid_height
        
        # Calculate primary direction
        dx = pos2[0] - pos1[0]
        dy = pos2[1] - pos1[1]
        
        # If segments are too far apart (wrapping), don't draw connecting segment
        if abs(dx) > 2 or abs(dy) > 2:
            return []
        
        # Convert to screen coordinates
        screen_pos1 = game_area.get_screen_pos(pos1[0], pos1[1])
        screen_pos2 = game_area.get_screen_pos(pos2[0], pos2[1])
        
        return [(screen_pos1, screen_pos2)]

