"""
Game Area Module

This module provides the GameArea class that handles the rendering of the game area
and coordinate conversions between screen and grid coordinates.
"""

import pygame
from src.config import (
    DEFAULT_PADDING,
    DEFAULT_SCORE_HEIGHT,
)
from src.config.colors import *
from typing import Tuple

class GameArea:
    """Manages the game area rendering and coordinate conversions."""
    
    def __init__(self, window_width, window_height, grid_size=20, grid_width=30, grid_height=30, 
                 padding=DEFAULT_PADDING, score_height=DEFAULT_SCORE_HEIGHT):
        # Grid properties
        self.grid_size = grid_size
        self.grid_width = grid_width 
        self.grid_height = grid_height

        # Area properties
        self.padding = padding
        self.score_height = score_height
        self.border_thickness = 4  # Increased border thickness

        # Calculate game area dimensions
        self.width = self.grid_width * self.grid_size
        self.height = self.grid_height * self.grid_size
        
        # Center the game area in the window
        self.x = (window_width - self.width) // 2
        self.y = self.score_height + self.padding
        
        # Create rectangle for game bounds
        self.rect = pygame.Rect(self.x, self.y, self.width, self.height)
        
        # Create inner rectangle for gameplay area with thicker border
        self.inner_rect = pygame.Rect(
            self.x + self.border_thickness,
            self.y + self.border_thickness,
            self.width - (self.border_thickness * 2),
            self.height - (self.border_thickness * 2)
        )
        
        self.screen_offset = (self.x, self.y)
    
    def get_grid_pos(self, screen_x, screen_y):
        """Convert screen coordinates to grid position."""
        grid_x = (screen_x - self.x) // self.grid_size
        grid_y = (screen_y - self.y) // self.grid_size
        return grid_x, grid_y
    
    def get_screen_pos(self, grid_x, grid_y):
        """Convert grid position to screen coordinates."""
        screen_x = self.x + (grid_x * self.grid_size)
        screen_y = self.y + (grid_y * self.grid_size)
        return screen_x, screen_y
    
    def is_within_bounds(self, grid_x, grid_y):
        """Check if grid position is within game bounds."""
        return 0 <= grid_x < self.grid_width and 0 <= grid_y < self.grid_height
    
    def draw(self, screen):
        """Draw the game area boundary and background with modern effects."""
        # Draw outer border with gradient effect
        for i in range(self.border_thickness):
            border_rect = pygame.Rect(
                self.x + i,
                self.y + i,
                self.width - (i * 2),
                self.height - (i * 2)
            )
            # Gradient from darker to lighter
            color_factor = i / self.border_thickness
            r = int(DARKER_GREEN[0] * (1 - color_factor) + FOREST_GREEN[0] * color_factor)
            g = int(DARKER_GREEN[1] * (1 - color_factor) + FOREST_GREEN[1] * color_factor)
            b = int(DARKER_GREEN[2] * (1 - color_factor) + FOREST_GREEN[2] * color_factor)
            pygame.draw.rect(screen, (r, g, b), border_rect)
        
        # Draw inner background
        pygame.draw.rect(screen, DARK_GRAY, self.inner_rect)
        
        # Draw subtle grid pattern
        if hasattr(self, 'show_grid') and self.show_grid:
            # Debug grid lines
            for x in range(self.grid_width + 1):
                start_pos = (self.x + x * self.grid_size, self.y)
                end_pos = (self.x + x * self.grid_size, self.y + self.height)
                pygame.draw.line(screen, GRID_COLOR, start_pos, end_pos, 1)
            
            for y in range(self.grid_height + 1):
                start_pos = (self.x, self.y + y * self.grid_size)
                end_pos = (self.x + self.width, self.y + y * self.grid_size)
                pygame.draw.line(screen, GRID_COLOR, start_pos, end_pos, 1)
        else:
            # Subtle grid dots with fade effect
            for x in range(self.grid_width + 1):
                for y in range(self.grid_height + 1):
                    pos_x = self.x + x * self.grid_size
                    pos_y = self.y + y * self.grid_size
                    
                    # Calculate distance from center for fade effect
                    center_x = self.x + (self.width / 2)
                    center_y = self.y + (self.height / 2)
                    dist_x = abs(pos_x - center_x) / (self.width / 2)
                    dist_y = abs(pos_y - center_y) / (self.height / 2)
                    dist = min(1.0, (dist_x + dist_y) / 2)
                    
                    # Fade dot color based on distance from center
                    dot_color = tuple(int(c * (1 - dist * 0.5)) for c in SUBTLE_GRID_COLOR)
                    pygame.draw.circle(screen, dot_color, (pos_x, pos_y), 1)
        
        # Draw inner border glow
        glow_colors = [
            (int(NEON_GREEN[0] * 0.4), int(NEON_GREEN[1] * 0.4), int(NEON_GREEN[2] * 0.4)),
            (int(NEON_GREEN[0] * 0.6), int(NEON_GREEN[1] * 0.6), int(NEON_GREEN[2] * 0.6)),
            NEON_GREEN
        ]
        
        for i, color in enumerate(glow_colors):
            glow_rect = pygame.Rect(
                self.inner_rect.x - i,
                self.inner_rect.y - i,
                self.inner_rect.width + (i * 2),
                self.inner_rect.height + (i * 2)
            )
            pygame.draw.rect(screen, color, glow_rect, 1)

    def grid_to_screen(self, grid_pos: Tuple[int, int]) -> Tuple[int, int]:
        """Convert grid coordinates to screen coordinates"""
        x = self.rect.left + grid_pos[0] * self.grid_size + self.grid_size // 2
        y = self.rect.top + grid_pos[1] * self.grid_size + self.grid_size // 2
        return (x, y) 