"""
Main Menu Module

This module provides the main menu interface for the game.
"""

import pygame
from enum import Enum, auto
from src.config import (
    WINDOW_WIDTH,
    WINDOW_HEIGHT,
    TITLE_FONT_SIZE,
    SUBTITLE_FONT_SIZE,
    MENU_ITEM_SPACING
)
from src.config.colors import BLACK, GRAY
from src.ui.menu_item import MenuItem

class GameMode(Enum):
    """Available game modes."""
    PLAYER = auto()
    AI_EASY = auto()
    AI_MEDIUM = auto()
    AI_HARD = auto()
    SETTINGS = auto()

class Menu:
    """Main menu interface."""
    
    def __init__(self, width=WINDOW_WIDTH, height=WINDOW_HEIGHT):
        """Initialize the main menu."""
        self.width = width
        self.height = height
        self.setup_menu_items()
        
        # Create fonts
        self.title_font = pygame.font.Font(None, TITLE_FONT_SIZE)
        self.subtitle_font = pygame.font.Font(None, SUBTITLE_FONT_SIZE)
        
        # Create title surfaces
        self.title_surface = self.title_font.render("Snake Game", True, (0, 255, 0))
        self.title_rect = self.title_surface.get_rect(center=(width//2, height//4))
        
        # Initialize first item as selected
        self.selected_index = 0
        self.menu_items[0].hover = True
        self.menu_items[0]._setup_font()
    
    def setup_menu_items(self):
        """Setup the menu items."""
        start_y = self.height // 2
        spacing = MENU_ITEM_SPACING
        center_x = self.width // 2
        
        self.menu_items = [
            MenuItem("Player Game", (center_x, start_y), GameMode.PLAYER),
            MenuItem("AI Game (Easy)", (center_x, start_y + spacing), GameMode.AI_EASY),
            MenuItem("AI Game (Medium)", (center_x, start_y + spacing * 2), GameMode.AI_MEDIUM),
            MenuItem("AI Game (Hard)", (center_x, start_y + spacing * 3), GameMode.AI_HARD),
            MenuItem("Settings", (center_x, start_y + spacing * 4), GameMode.SETTINGS),
            MenuItem("Quit", (center_x, start_y + spacing * 5), 'quit')
        ]
    
    def update(self):
        """Update menu state."""
        # Handle mouse hover
        mouse_pos = pygame.mouse.get_pos()
        for i, item in enumerate(self.menu_items):
            if item.rect.collidepoint(mouse_pos):
                # Update selected index when mouse hovers
                self.selected_index = i
                item.hover = True
                item._setup_font()
            else:
                # Keep keyboard selection visible
                item.hover = (i == self.selected_index)
                item._setup_font()
    
    def handle_input(self, event):
        """
        Handle input events.
        
        Returns:
            GameMode or None: The selected game mode or None if no selection made
        """
        if event.type == pygame.MOUSEBUTTONDOWN and event.button == 1:
            # Handle mouse clicks
            mouse_pos = pygame.mouse.get_pos()
            for item in self.menu_items:
                if item.rect.collidepoint(mouse_pos):
                    return item.action
        
        elif event.type == pygame.KEYDOWN:
            if event.key == pygame.K_RETURN:
                return self.menu_items[self.selected_index].action
            elif event.key in (pygame.K_UP, pygame.K_DOWN):
                # Update selected index
                if event.key == pygame.K_UP:
                    self.selected_index = (self.selected_index - 1) % len(self.menu_items)
                else:
                    self.selected_index = (self.selected_index + 1) % len(self.menu_items)
                
                # Update hover states
                for i, item in enumerate(self.menu_items):
                    item.hover = (i == self.selected_index)
                    item._setup_font()
        
        return None
    
    def draw(self, screen):
        """Draw the menu to the screen."""
        # Draw background
        screen.fill(BLACK)
        
        # Draw title
        screen.blit(self.title_surface, self.title_rect)
        
        # Draw menu items
        for item in self.menu_items:
            item.draw(screen)
        
        # Draw controls
        controls_text = "Arrow keys or mouse to navigate, Enter to select"
        controls_surface = self.subtitle_font.render(controls_text, True, GRAY)
        screen.blit(controls_surface, 
                   (self.width - controls_surface.get_width() - 10, 
                    self.height - 30)) 