"""
Menu Item Module

This module provides the MenuItem class that represents a clickable menu item.
"""

import pygame
from src.config.colors import *

class MenuItem:
    """A clickable menu item with hover effects."""
    
    def __init__(self, text, position, action, font_size_normal=36, font_size_hover=48):
        """
        Initialize a menu item.
        
        Args:
            text: The text to display
            position: (x, y) tuple for center position
            action: String identifier for the action to take when clicked
            font_size_normal: Font size when not hovered (default 36)
            font_size_hover: Font size when hovered (default 48)
        """
        self.text = text
        self.position = position
        self.action = action
        self.hover = False
        self.font_size_normal = font_size_normal
        self.font_size_hover = font_size_hover
        self._setup_font()
    
    def _setup_font(self):
        """Setup the font based on hover state."""
        size = self.font_size_hover if self.hover else self.font_size_normal
        self.font = pygame.font.Font(None, size)
        self.surface = self.font.render(self.text, True, GREEN if self.hover else WHITE)
        self.rect = self.surface.get_rect(center=self.position)
    
    def draw(self, screen):
        """Draw the menu item to the screen."""
        screen.blit(self.surface, self.rect)


class NumberMenuItem(MenuItem):
    """A menu item that displays a number that can be adjusted with arrow keys."""
    
    def __init__(self, text, position, action, min_value, max_value, 
                 step=1, font_size_normal=36, font_size_hover=48):
        """
        Initialize a number menu item.
        
        Args:
            text: The text to display
            position: (x, y) tuple for center position  
            action: String identifier for the action to take when clicked
            min_value: Minimum allowed value
            max_value: Maximum allowed value
            step: Amount to increment/decrement by (default 1)
            font_size_normal: Font size when not hovered (default 36)
            font_size_hover: Font size when hovered (default 48)
        """
        super().__init__(text, position, action, font_size_normal, font_size_hover)
        self.min_value = min_value
        self.max_value = max_value
        self.step = step
        self.current_value = int(text.split(": ")[1])  # Extract number from text
    
    def increment(self):
        """Increment the current value by step amount."""
        if self.current_value + self.step <= self.max_value:
            self.current_value += self.step
            self._update_text()
    
    def decrement(self):
        """Decrement the current value by step amount."""
        if self.current_value - self.step >= self.min_value:
            self.current_value -= self.step
            self._update_text()
            
    def _update_text(self):
        """Update the displayed text with new value."""
        base_text = self.text.split(": ")[0]
        self.text = f"{base_text}: {self.current_value}"
        self._setup_font()
