"""
Settings Menu Module

This module provides the settings menu interface for configuring game rules.
"""

import pygame
from src.config import (
    WINDOW_WIDTH,
    WINDOW_HEIGHT,
    TITLE_FONT_SIZE,
    SUBTITLE_FONT_SIZE,
    MENU_ITEM_SPACING,
)
from src.config.colors import *
from src.ui.menu_item import MenuItem, NumberMenuItem

class SettingsMenu:
    """Settings menu interface."""
    
    def __init__(self, width=WINDOW_WIDTH, height=WINDOW_HEIGHT, rules=None):
        """
        Initialize the settings menu.
        
        Args:
            width: Window width
            height: Window height
            rules: GameRules instance to modify
        """
        self.width = width
        self.height = height
        self.rules = rules
        self.setup_menu_items()
        
        self.title_font = pygame.font.Font(None, TITLE_FONT_SIZE)
        self.subtitle_font = pygame.font.Font(None, SUBTITLE_FONT_SIZE)
        
        # Create title surfaces
        self.title_surface = self.title_font.render("Settings", True, (0, 255, 0))
        self.title_rect = self.title_surface.get_rect(center=(width//2, height//4))
        
        # Initialize first item as selected
        self.selected_index = 0
        self.menu_items[0].hover = True
        self.menu_items[0]._setup_font()
    
    def setup_menu_items(self):
        """Setup the menu items."""
        start_y = self.height // 2
        spacing = MENU_ITEM_SPACING
        center_x = self.width // 2
        
        self.menu_items = [
            MenuItem(f"Wrap Around: {'On' if self.rules.wrap_around else 'Off'}", 
                    (center_x, start_y), 
                    'toggle_wrap'),
            MenuItem(f"Speed Increase: {'On' if self.rules.speed_increase else 'Off'}", 
                    (center_x, start_y + spacing), 
                    'toggle_speed'),
            NumberMenuItem(f"Starting Length: {self.rules.starting_length}", 
                    (center_x, start_y + spacing * 2), 
                    'set_starting_length', 
                    1, 10),
            MenuItem("Back to Menu", 
                    (center_x, start_y + spacing * 3), 
                    'back')
        ]
    
    def update(self):
        """Update menu state."""
        # Handle mouse hover
        mouse_pos = pygame.mouse.get_pos()
        for i, item in enumerate(self.menu_items):
            if item.rect.collidepoint(mouse_pos):
                # Update selected index when mouse hovers
                self.selected_index = i
                item.hover = True
                item._setup_font()
            else:
                # Keep keyboard selection visible
                item.hover = (i == self.selected_index)
                item._setup_font()
    
    def handle_input(self, event):
        """
        Handle input events.
        
        Returns:
            str or None: The action to take or None if no action
        """
        if event.type == pygame.MOUSEBUTTONDOWN:
            # Handle mouse clicks
            mouse_pos = pygame.mouse.get_pos()
            for i, item in enumerate(self.menu_items):
                if item.rect.collidepoint(mouse_pos):
                    self.selected_index = i
                    if event.button == 1:
                        if item.action == 'toggle_wrap':
                            self.rules.update_rule('wrap_around', not self.rules.wrap_around)
                        elif item.action == 'toggle_speed':
                            self.rules.speed_increase = not self.rules.speed_increase
                            item.text = f"Speed Increase: {'On' if self.rules.speed_increase else 'Off'}"
                            item._setup_font()
                        elif item.action == 'set_starting_length':
                            item.increment()
                            self.rules.update_rule('starting_length', item.current_value)
                        elif item.action == 'back':
                            return 'back'
                    elif event.button == 3:
                        if item.action == 'set_starting_length':
                            item.decrement()
                            self.rules.update_rule('starting_length', item.current_value)

        elif event.type == pygame.KEYDOWN:
            if event.key == pygame.K_ESCAPE:
                return 'back'
            elif event.key == pygame.K_RETURN:
                item = self.menu_items[self.selected_index]
                if item.action == 'toggle_wrap':
                    self.rules.update_rule('wrap_around', not self.rules.wrap_around)
                elif item.action == 'toggle_speed':
                    self.rules.update_rule('speed_increase', not self.rules.speed_increase)
                elif item.action == 'set_starting_length':
                    item.increment()
                    self.rules.update_rule('starting_length', item.current_value)
                elif item.action == 'back':
                    return 'back'
            elif event.key in (pygame.K_UP, pygame.K_DOWN):
                # Update selected index
                if event.key == pygame.K_UP:
                    self.selected_index = (self.selected_index - 1) % len(self.menu_items)
                else:
                    self.selected_index = (self.selected_index + 1) % len(self.menu_items)
                
                # Update hover states
                for i, item in enumerate(self.menu_items):
                    item.hover = (i == self.selected_index)
                    item._setup_font()
        
        return None
    
    def draw(self, screen):
        """Draw the menu to the screen."""
        # Draw background
        screen.fill(BLACK)
        
        # Draw title
        screen.blit(self.title_surface, self.title_rect)
        
        # Draw menu items
        for item in self.menu_items:
            item.draw(screen)
        
        # Draw controls
        controls_text = "Arrow keys or mouse to navigate, Enter to select, Esc to go back"
        controls_surface = self.subtitle_font.render(controls_text, True, GRAY)
        screen.blit(controls_surface, 
                   (self.width - controls_surface.get_width() - 10, 
                    self.height - 30)) 